# This file is part of beets.
# Copyright 2016, Adrian Sampson.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Tests for MusicBrainz API wrapper."""

import unittest
import uuid
from typing import Any, ClassVar
from unittest import mock

import pytest
import requests

from beets import config
from beets.library import Item
from beets.test.helper import BeetsTestCase, PluginMixin
from beetsplug import musicbrainz


def make_alias(suffix: str, locale: str, primary: bool = False):
    alias: dict[str, Any] = {
        "name": f"ALIAS{suffix}",
        "locale": locale,
        "sort-name": f"ALIASSORT{suffix}",
    }
    if primary:
        alias["primary"] = True
    return alias


class MusicBrainzTestCase(BeetsTestCase):
    def setUp(self):
        super().setUp()
        self.mb = musicbrainz.MusicBrainzPlugin()
        self.config["match"]["preferred"]["countries"] = ["US"]


class MBAlbumInfoTest(MusicBrainzTestCase):
    def _make_release(
        self,
        date_str="2009",
        tracks=None,
        track_length=None,
        track_artist=False,
        multi_artist_credit=False,
        data_tracks=None,
        medium_format="FORMAT",
    ):
        release = {
            "title": "ALBUM TITLE",
            "id": "ALBUM ID",
            "asin": "ALBUM ASIN",
            "disambiguation": "R_DISAMBIGUATION",
            "release-group": {
                "primary-type": "Album",
                "first-release-date": date_str,
                "id": "RELEASE GROUP ID",
                "disambiguation": "RG_DISAMBIGUATION",
                "title": "RELEASE GROUP TITLE",
            },
            "artist-credit": [
                {
                    "artist": {
                        "name": "ARTIST NAME",
                        "id": "ARTIST ID",
                        "sort-name": "ARTIST SORT NAME",
                    },
                    "name": "ARTIST CREDIT",
                }
            ],
            "date": "3001",
            "media": [],
            "genres": [{"count": 1, "name": "GENRE"}],
            "tags": [{"count": 1, "name": "TAG"}],
            "label-info": [
                {
                    "catalog-number": "CATALOG NUMBER",
                    "label": {"name": "LABEL NAME"},
                }
            ],
            "text-representation": {
                "script": "SCRIPT",
                "language": "LANGUAGE",
            },
            "country": "COUNTRY",
            "status": "STATUS",
            "barcode": "BARCODE",
            "release-events": [{"area": None, "date": "2021-03-26"}],
        }

        if multi_artist_credit:
            release["artist-credit"][0]["joinphrase"] = " & "
            release["artist-credit"].append(
                {
                    "artist": {
                        "name": "ARTIST 2 NAME",
                        "id": "ARTIST 2 ID",
                        "sort-name": "ARTIST 2 SORT NAME",
                    },
                    "name": "ARTIST MULTI CREDIT",
                }
            )

        i = 0
        track_list = []
        if tracks:
            for recording in tracks:
                i += 1
                track = {
                    "id": f"RELEASE TRACK ID {i}",
                    "recording": recording,
                    "position": i,
                    "number": "A1",
                }
                if track_length:
                    # Track lengths are distinct from recording lengths.
                    track["length"] = track_length
                if track_artist:
                    # Similarly, track artists can differ from recording
                    # artists.
                    track["artist-credit"] = [
                        {
                            "artist": {
                                "name": "TRACK ARTIST NAME",
                                "id": "TRACK ARTIST ID",
                                "sort-name": "TRACK ARTIST SORT NAME",
                            },
                            "name": "TRACK ARTIST CREDIT",
                        }
                    ]

                    if multi_artist_credit:
                        track["artist-credit"][0]["joinphrase"] = " & "
                        track["artist-credit"].append(
                            {
                                "artist": {
                                    "name": "TRACK ARTIST 2 NAME",
                                    "id": "TRACK ARTIST 2 ID",
                                    "sort-name": "TRACK ARTIST 2 SORT NAME",
                                },
                                "name": "TRACK ARTIST 2 CREDIT",
                            }
                        )

                track_list.append(track)
        data_track_list = []
        if data_tracks:
            for recording in data_tracks:
                i += 1
                data_track = {
                    "id": f"RELEASE TRACK ID {i}",
                    "recording": recording,
                    "position": i,
                    "number": "A1",
                }
                data_track_list.append(data_track)
        release["media"].append(
            {
                "position": "1",
                "tracks": track_list,
                "data-tracks": data_track_list,
                "format": medium_format,
                "title": "MEDIUM TITLE",
            }
        )
        return release

    def _make_track(
        self,
        title,
        tr_id,
        duration,
        artist=False,
        video=False,
        disambiguation=None,
        remixer=False,
        multi_artist_credit=False,
        aliases=None,
    ):
        track = {
            "title": title,
            "id": tr_id,
        }
        if duration is not None:
            track["length"] = duration
        if artist:
            track["artist-credit"] = [
                {
                    "artist": {
                        "name": "RECORDING ARTIST NAME",
                        "id": "RECORDING ARTIST ID",
                        "sort-name": "RECORDING ARTIST SORT NAME",
                    },
                    "name": "RECORDING ARTIST CREDIT",
                }
            ]
            if multi_artist_credit:
                track["artist-credit"][0]["joinphrase"] = " & "
                track["artist-credit"].append(
                    {
                        "artist": {
                            "name": "RECORDING ARTIST 2 NAME",
                            "id": "RECORDING ARTIST 2 ID",
                            "sort-name": "RECORDING ARTIST 2 SORT NAME",
                        },
                        "name": "RECORDING ARTIST 2 CREDIT",
                    }
                )
        if remixer:
            track["artist-relations"] = [
                {
                    "type": "remixer",
                    "type-id": "RELATION TYPE ID",
                    "direction": "RECORDING RELATION DIRECTION",
                    "artist": {
                        "id": "RECORDING REMIXER ARTIST ID",
                        "type": "RECORDING REMIXER ARTIST TYPE",
                        "name": "RECORDING REMIXER ARTIST NAME",
                        "sort-name": "RECORDING REMIXER ARTIST SORT NAME",
                    },
                }
            ]
        if video:
            track["video"] = True
        if disambiguation:
            track["disambiguation"] = disambiguation
        if aliases is not None:
            track["aliases"] = aliases
        return track

    def test_parse_release_title(self):
        release = self._make_release(None)
        release["aliases"] = [
            make_alias(suffix="en", locale="en", primary=True),
        ]

        # test no alias
        config["import"]["languages"] = []
        d = self.mb.album_info(release)
        assert d.album == "ALBUM TITLE"

        # test en primary
        config["import"]["languages"] = ["en"]
        d = self.mb.album_info(release)
        assert d.album == "ALIASen"

    def test_parse_release_with_year(self):
        release = self._make_release("1984")
        d = self.mb.album_info(release)
        assert d.album == "ALBUM TITLE"
        assert d.album_id == "ALBUM ID"
        assert d.artist == "ARTIST NAME"
        assert d.artist_id == "ARTIST ID"
        assert d.original_year == 1984
        assert d.year == 3001
        assert d.artist_credit == "ARTIST CREDIT"

    def test_parse_release_type(self):
        release = self._make_release("1984")
        d = self.mb.album_info(release)
        assert d.albumtype == "album"

    def test_parse_release_full_date(self):
        release = self._make_release("1987-03-31")
        d = self.mb.album_info(release)
        assert d.original_year == 1987
        assert d.original_month == 3
        assert d.original_day == 31

    def test_parse_tracks(self):
        tracks = [
            self._make_track(
                "TITLE ONE",
                "ID ONE",
                100.0 * 1000.0,
                aliases=[make_alias(suffix="ONEen", locale="en", primary=True)],
            ),
            self._make_track(
                "TITLE TWO",
                "ID TWO",
                200.0 * 1000.0,
                aliases=[make_alias(suffix="TWOen", locale="en", primary=True)],
            ),
        ]
        release = self._make_release(tracks=tracks)

        # Preference over recording data
        release["media"][0]["tracks"][1]["title"] = "TRACK TITLE TWO"

        # test no alias
        config["import"]["languages"] = []
        d = self.mb.album_info(release)
        t = d.tracks
        assert len(t) == 2
        assert t[0].title == "TITLE ONE"
        assert t[0].track_id == "ID ONE"
        assert t[0].length == 100.0
        assert t[1].title == "TRACK TITLE TWO"
        assert t[1].track_id == "ID TWO"
        assert t[1].length == 200.0

        # test en primary
        config["import"]["languages"] = ["en"]
        d = self.mb.album_info(release)
        t = d.tracks
        assert len(t) == 2
        assert t[0].title == "ALIASONEen"
        assert t[0].track_id == "ID ONE"
        assert t[0].length == 100.0
        assert t[1].title == "ALIASTWOen"
        assert t[1].track_id == "ID TWO"
        assert t[1].length == 200.0

    def test_parse_track_indices(self):
        tracks = [
            self._make_track("TITLE ONE", "ID ONE", 100.0 * 1000.0),
            self._make_track("TITLE TWO", "ID TWO", 200.0 * 1000.0),
        ]
        release = self._make_release(tracks=tracks)

        d = self.mb.album_info(release)
        t = d.tracks
        assert t[0].medium_index == 1
        assert t[0].index == 1
        assert t[1].medium_index == 2
        assert t[1].index == 2

    def test_parse_medium_numbers_single_medium(self):
        tracks = [
            self._make_track("TITLE ONE", "ID ONE", 100.0 * 1000.0),
            self._make_track("TITLE TWO", "ID TWO", 200.0 * 1000.0),
        ]
        release = self._make_release(tracks=tracks)

        d = self.mb.album_info(release)
        assert d.mediums == 1
        t = d.tracks
        assert t[0].medium == 1
        assert t[1].medium == 1

    def test_parse_medium_numbers_two_mediums(self):
        tracks = [
            self._make_track("TITLE ONE", "ID ONE", 100.0 * 1000.0),
            self._make_track("TITLE TWO", "ID TWO", 200.0 * 1000.0),
        ]
        release = self._make_release(tracks=[tracks[0]])
        second_track_list = [
            {
                "id": "RELEASE TRACK ID 2",
                "recording": tracks[1],
                "position": "1",
                "number": "A1",
            }
        ]
        release["media"].append(
            {
                "position": "2",
                "tracks": second_track_list,
            }
        )

        d = self.mb.album_info(release)
        assert d.mediums == 2
        t = d.tracks
        assert t[0].medium == 1
        assert t[0].medium_index == 1
        assert t[0].index == 1
        assert t[1].medium == 2
        assert t[1].medium_index == 1
        assert t[1].index == 2

    def test_parse_release_year_month_only(self):
        release = self._make_release("1987-03")
        d = self.mb.album_info(release)
        assert d.original_year == 1987
        assert d.original_month == 3

    def test_no_durations(self):
        tracks = [self._make_track("TITLE", "ID", None)]
        release = self._make_release(tracks=tracks)
        d = self.mb.album_info(release)
        assert d.tracks[0].length is None

    def test_track_length_overrides_recording_length(self):
        tracks = [self._make_track("TITLE", "ID", 1.0 * 1000.0)]
        release = self._make_release(tracks=tracks, track_length=2.0 * 1000.0)
        d = self.mb.album_info(release)
        assert d.tracks[0].length == 2.0

    def test_no_release_date(self):
        release = self._make_release(None)
        d = self.mb.album_info(release)
        assert not d.original_year
        assert not d.original_month
        assert not d.original_day

    def test_various_artists_defaults_false(self):
        release = self._make_release(None)
        d = self.mb.album_info(release)
        assert not d.va

    def test_detect_various_artists(self):
        release = self._make_release(None)
        release["artist-credit"][0]["artist"]["id"] = (
            musicbrainz.VARIOUS_ARTISTS_ID
        )
        d = self.mb.album_info(release)
        assert d.va

    def test_parse_artist_sort_name(self):
        release = self._make_release(None)
        d = self.mb.album_info(release)
        assert d.artist_sort == "ARTIST SORT NAME"

    def test_parse_releasegroupid(self):
        release = self._make_release(None)
        d = self.mb.album_info(release)
        assert d.releasegroup_id == "RELEASE GROUP ID"

    def test_parse_release_group_title(self):
        release = self._make_release(None)
        release["release-group"]["aliases"] = [
            make_alias(suffix="en", locale="en", primary=True),
        ]

        # test no alias
        config["import"]["languages"] = []
        d = self.mb.album_info(release)
        assert d.release_group_title == "RELEASE GROUP TITLE"

        # test en primary
        config["import"]["languages"] = ["en"]
        d = self.mb.album_info(release)
        assert d.release_group_title == "ALIASen"

    def test_parse_asin(self):
        release = self._make_release(None)
        d = self.mb.album_info(release)
        assert d.asin == "ALBUM ASIN"

    def test_parse_catalognum(self):
        release = self._make_release(None)
        d = self.mb.album_info(release)
        assert d.catalognum == "CATALOG NUMBER"

    def test_parse_textrepr(self):
        release = self._make_release(None)
        d = self.mb.album_info(release)
        assert d.script == "SCRIPT"
        assert d.language == "LANGUAGE"

    def test_parse_country(self):
        release = self._make_release(None)
        d = self.mb.album_info(release)
        assert d.country == "COUNTRY"

    def test_parse_status(self):
        release = self._make_release(None)
        d = self.mb.album_info(release)
        assert d.albumstatus == "STATUS"

    def test_parse_barcode(self):
        release = self._make_release(None)
        d = self.mb.album_info(release)
        assert d.barcode == "BARCODE"

    def test_parse_media(self):
        tracks = [
            self._make_track("TITLE ONE", "ID ONE", 100.0 * 1000.0),
            self._make_track("TITLE TWO", "ID TWO", 200.0 * 1000.0),
        ]
        release = self._make_release(None, tracks=tracks)
        d = self.mb.album_info(release)
        assert d.media == "FORMAT"

    def test_parse_disambig(self):
        release = self._make_release(None)
        d = self.mb.album_info(release)
        assert d.albumdisambig == "R_DISAMBIGUATION"
        assert d.releasegroupdisambig == "RG_DISAMBIGUATION"

    def test_parse_disctitle(self):
        tracks = [
            self._make_track("TITLE ONE", "ID ONE", 100.0 * 1000.0),
            self._make_track("TITLE TWO", "ID TWO", 200.0 * 1000.0),
        ]
        release = self._make_release(None, tracks=tracks)
        d = self.mb.album_info(release)
        t = d.tracks
        assert t[0].disctitle == "MEDIUM TITLE"
        assert t[1].disctitle == "MEDIUM TITLE"

    def test_missing_language(self):
        release = self._make_release(None)
        del release["text-representation"]["language"]
        d = self.mb.album_info(release)
        assert d.language is None

    def test_parse_recording_artist(self):
        tracks = [self._make_track("a", "b", 1, True)]
        release = self._make_release(None, tracks=tracks)
        track = self.mb.album_info(release).tracks[0]
        assert track.artist == "RECORDING ARTIST NAME"
        assert track.artist_id == "RECORDING ARTIST ID"
        assert track.artist_sort == "RECORDING ARTIST SORT NAME"
        assert track.artist_credit == "RECORDING ARTIST CREDIT"

    def test_parse_recording_artist_multi(self):
        tracks = [self._make_track("a", "b", 1, True, multi_artist_credit=True)]
        release = self._make_release(None, tracks=tracks)
        track = self.mb.album_info(release).tracks[0]
        assert track.artist == "RECORDING ARTIST NAME & RECORDING ARTIST 2 NAME"
        assert track.artist_id == "RECORDING ARTIST ID"
        assert (
            track.artist_sort
            == "RECORDING ARTIST SORT NAME & RECORDING ARTIST 2 SORT NAME"
        )
        assert (
            track.artist_credit
            == "RECORDING ARTIST CREDIT & RECORDING ARTIST 2 CREDIT"
        )

        assert track.artists == [
            "RECORDING ARTIST NAME",
            "RECORDING ARTIST 2 NAME",
        ]
        assert track.artists_ids == [
            "RECORDING ARTIST ID",
            "RECORDING ARTIST 2 ID",
        ]
        assert track.artists_sort == [
            "RECORDING ARTIST SORT NAME",
            "RECORDING ARTIST 2 SORT NAME",
        ]
        assert track.artists_credit == [
            "RECORDING ARTIST CREDIT",
            "RECORDING ARTIST 2 CREDIT",
        ]

    def test_track_artist_overrides_recording_artist(self):
        tracks = [self._make_track("a", "b", 1, True)]
        release = self._make_release(None, tracks=tracks, track_artist=True)
        track = self.mb.album_info(release).tracks[0]
        assert track.artist == "TRACK ARTIST NAME"
        assert track.artist_id == "TRACK ARTIST ID"
        assert track.artist_sort == "TRACK ARTIST SORT NAME"
        assert track.artist_credit == "TRACK ARTIST CREDIT"

    def test_track_artist_overrides_recording_artist_multi(self):
        tracks = [self._make_track("a", "b", 1, True, multi_artist_credit=True)]
        release = self._make_release(
            None, tracks=tracks, track_artist=True, multi_artist_credit=True
        )
        track = self.mb.album_info(release).tracks[0]
        assert track.artist == "TRACK ARTIST NAME & TRACK ARTIST 2 NAME"
        assert track.artist_id == "TRACK ARTIST ID"
        assert (
            track.artist_sort
            == "TRACK ARTIST SORT NAME & TRACK ARTIST 2 SORT NAME"
        )
        assert (
            track.artist_credit == "TRACK ARTIST CREDIT & TRACK ARTIST 2 CREDIT"
        )

        assert track.artists == ["TRACK ARTIST NAME", "TRACK ARTIST 2 NAME"]
        assert track.artists_ids == ["TRACK ARTIST ID", "TRACK ARTIST 2 ID"]
        assert track.artists_sort == [
            "TRACK ARTIST SORT NAME",
            "TRACK ARTIST 2 SORT NAME",
        ]
        assert track.artists_credit == [
            "TRACK ARTIST CREDIT",
            "TRACK ARTIST 2 CREDIT",
        ]

    def test_parse_recording_remixer(self):
        tracks = [self._make_track("a", "b", 1, remixer=True)]
        release = self._make_release(None, tracks=tracks)
        track = self.mb.album_info(release).tracks[0]
        assert track.remixer == "RECORDING REMIXER ARTIST NAME"

    def test_data_source(self):
        release = self._make_release()
        d = self.mb.album_info(release)
        assert d.data_source == "MusicBrainz"

    def test_genres(self):
        config["musicbrainz"]["genres"] = True
        config["musicbrainz"]["genres_tag"] = "genre"
        release = self._make_release()
        d = self.mb.album_info(release)
        assert d.genres == ["GENRE"]

    def test_tags(self):
        config["musicbrainz"]["genres"] = True
        config["musicbrainz"]["genres_tag"] = "tag"
        release = self._make_release()
        d = self.mb.album_info(release)
        assert d.genres == ["TAG"]

    def test_no_genres(self):
        config["musicbrainz"]["genres"] = False
        release = self._make_release()
        d = self.mb.album_info(release)
        assert d.genres is None

    def test_ignored_media(self):
        config["match"]["ignored_media"] = ["IGNORED1", "IGNORED2"]
        tracks = [
            self._make_track("TITLE ONE", "ID ONE", 100.0 * 1000.0),
            self._make_track("TITLE TWO", "ID TWO", 200.0 * 1000.0),
        ]
        release = self._make_release(tracks=tracks, medium_format="IGNORED1")
        d = self.mb.album_info(release)
        assert len(d.tracks) == 0

    def test_no_ignored_media(self):
        config["match"]["ignored_media"] = ["IGNORED1", "IGNORED2"]
        tracks = [
            self._make_track("TITLE ONE", "ID ONE", 100.0 * 1000.0),
            self._make_track("TITLE TWO", "ID TWO", 200.0 * 1000.0),
        ]
        release = self._make_release(tracks=tracks, medium_format="NON-IGNORED")
        d = self.mb.album_info(release)
        assert len(d.tracks) == 2

    def test_skip_data_track(self):
        tracks = [
            self._make_track("TITLE ONE", "ID ONE", 100.0 * 1000.0),
            self._make_track("[data track]", "ID DATA TRACK", 100.0 * 1000.0),
            self._make_track("TITLE TWO", "ID TWO", 200.0 * 1000.0),
        ]
        release = self._make_release(tracks=tracks)
        d = self.mb.album_info(release)
        assert len(d.tracks) == 2
        assert d.tracks[0].title == "TITLE ONE"
        assert d.tracks[1].title == "TITLE TWO"

    def test_skip_audio_data_tracks_by_default(self):
        tracks = [
            self._make_track("TITLE ONE", "ID ONE", 100.0 * 1000.0),
            self._make_track("TITLE TWO", "ID TWO", 200.0 * 1000.0),
        ]
        data_tracks = [
            self._make_track(
                "TITLE AUDIO DATA", "ID DATA TRACK", 100.0 * 1000.0
            )
        ]
        release = self._make_release(tracks=tracks, data_tracks=data_tracks)
        d = self.mb.album_info(release)
        assert len(d.tracks) == 2
        assert d.tracks[0].title == "TITLE ONE"
        assert d.tracks[1].title == "TITLE TWO"

    def test_no_skip_audio_data_tracks_if_configured(self):
        config["match"]["ignore_data_tracks"] = False
        tracks = [
            self._make_track("TITLE ONE", "ID ONE", 100.0 * 1000.0),
            self._make_track("TITLE TWO", "ID TWO", 200.0 * 1000.0),
        ]
        data_tracks = [
            self._make_track(
                "TITLE AUDIO DATA", "ID DATA TRACK", 100.0 * 1000.0
            )
        ]
        release = self._make_release(tracks=tracks, data_tracks=data_tracks)
        d = self.mb.album_info(release)
        assert len(d.tracks) == 3
        assert d.tracks[0].title == "TITLE ONE"
        assert d.tracks[1].title == "TITLE TWO"
        assert d.tracks[2].title == "TITLE AUDIO DATA"

    def test_skip_video_tracks_by_default(self):
        tracks = [
            self._make_track("TITLE ONE", "ID ONE", 100.0 * 1000.0),
            self._make_track(
                "TITLE VIDEO", "ID VIDEO", 100.0 * 1000.0, False, True
            ),
            self._make_track("TITLE TWO", "ID TWO", 200.0 * 1000.0),
        ]
        release = self._make_release(tracks=tracks)
        d = self.mb.album_info(release)
        assert len(d.tracks) == 2
        assert d.tracks[0].title == "TITLE ONE"
        assert d.tracks[1].title == "TITLE TWO"

    def test_skip_video_data_tracks_by_default(self):
        tracks = [
            self._make_track("TITLE ONE", "ID ONE", 100.0 * 1000.0),
            self._make_track("TITLE TWO", "ID TWO", 200.0 * 1000.0),
        ]
        data_tracks = [
            self._make_track(
                "TITLE VIDEO", "ID VIDEO", 100.0 * 1000.0, False, True
            )
        ]
        release = self._make_release(tracks=tracks, data_tracks=data_tracks)
        d = self.mb.album_info(release)
        assert len(d.tracks) == 2
        assert d.tracks[0].title == "TITLE ONE"
        assert d.tracks[1].title == "TITLE TWO"

    def test_no_skip_video_tracks_if_configured(self):
        config["match"]["ignore_data_tracks"] = False
        config["match"]["ignore_video_tracks"] = False
        tracks = [
            self._make_track("TITLE ONE", "ID ONE", 100.0 * 1000.0),
            self._make_track(
                "TITLE VIDEO", "ID VIDEO", 100.0 * 1000.0, False, True
            ),
            self._make_track("TITLE TWO", "ID TWO", 200.0 * 1000.0),
        ]
        release = self._make_release(tracks=tracks)
        d = self.mb.album_info(release)
        assert len(d.tracks) == 3
        assert d.tracks[0].title == "TITLE ONE"
        assert d.tracks[1].title == "TITLE VIDEO"
        assert d.tracks[2].title == "TITLE TWO"

    def test_no_skip_video_data_tracks_if_configured(self):
        config["match"]["ignore_data_tracks"] = False
        config["match"]["ignore_video_tracks"] = False
        tracks = [
            self._make_track("TITLE ONE", "ID ONE", 100.0 * 1000.0),
            self._make_track("TITLE TWO", "ID TWO", 200.0 * 1000.0),
        ]
        data_tracks = [
            self._make_track(
                "TITLE VIDEO", "ID VIDEO", 100.0 * 1000.0, False, True
            )
        ]
        release = self._make_release(tracks=tracks, data_tracks=data_tracks)
        d = self.mb.album_info(release)
        assert len(d.tracks) == 3
        assert d.tracks[0].title == "TITLE ONE"
        assert d.tracks[1].title == "TITLE TWO"
        assert d.tracks[2].title == "TITLE VIDEO"

    def test_track_disambiguation(self):
        tracks = [
            self._make_track("TITLE ONE", "ID ONE", 100.0 * 1000.0),
            self._make_track(
                "TITLE TWO",
                "ID TWO",
                200.0 * 1000.0,
                disambiguation="SECOND TRACK",
            ),
        ]
        release = self._make_release(tracks=tracks)

        d = self.mb.album_info(release)
        t = d.tracks
        assert len(t) == 2
        assert t[0].trackdisambig is None
        assert t[1].trackdisambig == "SECOND TRACK"

    def test_missing_tracks(self):
        tracks = [
            self._make_track("TITLE ONE", "ID ONE", 100.0 * 1000.0),
            self._make_track(
                "TITLE TWO",
                "ID TWO",
                200.0 * 1000.0,
                disambiguation="SECOND TRACK",
            ),
        ]
        release = self._make_release(tracks=tracks)
        release["media"].append(release["media"][0])
        del release["media"][0]["tracks"]
        del release["media"][0]["data-tracks"]
        d = self.mb.album_info(release)
        assert d.mediums == 2


class ArtistFlatteningTest(unittest.TestCase):
    def _credit_dict(self, suffix=""):
        return {
            "artist": {
                "name": f"NAME{suffix}",
                "sort-name": f"SORT{suffix}",
            },
            "name": f"CREDIT{suffix}",
        }

    def test_single_artist(self):
        credit = [self._credit_dict()]
        a, s, c = musicbrainz._flatten_artist_credit(credit)
        assert a == "NAME"
        assert s == "SORT"
        assert c == "CREDIT"

        a, s, c = musicbrainz._multi_artist_credit(
            credit, include_join_phrase=False
        )
        assert a == ["NAME"]
        assert s == ["SORT"]
        assert c == ["CREDIT"]

    def test_two_artists(self):
        credit = [
            {**self._credit_dict("a"), "joinphrase": " AND "},
            self._credit_dict("b"),
        ]
        a, s, c = musicbrainz._flatten_artist_credit(credit)
        assert a == "NAMEa AND NAMEb"
        assert s == "SORTa AND SORTb"
        assert c == "CREDITa AND CREDITb"

        a, s, c = musicbrainz._multi_artist_credit(
            credit, include_join_phrase=False
        )
        assert a == ["NAMEa", "NAMEb"]
        assert s == ["SORTa", "SORTb"]
        assert c == ["CREDITa", "CREDITb"]

    def test_alias(self):
        credit_dict = self._credit_dict()
        credit_dict["artist"]["aliases"] = [
            make_alias(suffix="en", locale="en", primary=True),
            make_alias(suffix="en_GB", locale="en_GB", primary=True),
            make_alias(suffix="fr", locale="fr"),
            make_alias(suffix="fr_P", locale="fr", primary=True),
            make_alias(suffix="pt_BR", locale="pt_BR"),
        ]
        # test no alias
        config["import"]["languages"] = []
        flat = musicbrainz._flatten_artist_credit([credit_dict])
        assert flat == ("NAME", "SORT", "CREDIT")

        # test en primary
        config["import"]["languages"] = ["en"]
        flat = musicbrainz._flatten_artist_credit([credit_dict])
        assert flat == ("ALIASen", "ALIASSORTen", "CREDIT")

        # test en_GB en primary
        config["import"]["languages"] = ["en_GB", "en"]
        flat = musicbrainz._flatten_artist_credit([credit_dict])
        assert flat == ("ALIASen_GB", "ALIASSORTen_GB", "CREDIT")

        # test en en_GB primary
        config["import"]["languages"] = ["en", "en_GB"]
        flat = musicbrainz._flatten_artist_credit([credit_dict])
        assert flat == ("ALIASen", "ALIASSORTen", "CREDIT")

        # test fr primary
        config["import"]["languages"] = ["fr"]
        flat = musicbrainz._flatten_artist_credit([credit_dict])
        assert flat == ("ALIASfr_P", "ALIASSORTfr_P", "CREDIT")

        # test for not matching non-primary
        config["import"]["languages"] = ["pt_BR", "fr"]
        flat = musicbrainz._flatten_artist_credit([credit_dict])
        assert flat == ("ALIASfr_P", "ALIASSORTfr_P", "CREDIT")


class MBLibraryTest(MusicBrainzTestCase):
    def test_follow_pseudo_releases(self):
        side_effect = [
            {
                "title": "pseudo",
                "id": "d2a6f856-b553-40a0-ac54-a321e8e2da02",
                "status": "Pseudo-Release",
                "media": [
                    {
                        "tracks": [
                            {
                                "id": "baz",
                                "recording": {
                                    "title": "translated title",
                                    "id": "bar",
                                    "length": 42,
                                },
                                "position": 9,
                                "number": "A1",
                            }
                        ],
                        "position": 5,
                    }
                ],
                "artist-credit": [
                    {
                        "artist": {
                            "name": "some-artist",
                            "id": "some-id",
                        },
                    }
                ],
                "release-group": {
                    "id": "another-id",
                },
                "release-relations": [
                    {
                        "type": "transl-tracklisting",
                        "direction": "backward",
                        "release": {
                            "id": "d2a6f856-b553-40a0-ac54-a321e8e2da01"
                        },
                    }
                ],
            },
            {
                "title": "actual",
                "id": "d2a6f856-b553-40a0-ac54-a321e8e2da01",
                "status": "Official",
                "media": [
                    {
                        "tracks": [
                            {
                                "id": "baz",
                                "recording": {
                                    "title": "original title",
                                    "id": "bar",
                                    "length": 42,
                                },
                                "position": 9,
                                "number": "A1",
                            }
                        ],
                        "position": 5,
                    }
                ],
                "artist-credit": [
                    {
                        "artist": {
                            "name": "some-artist",
                            "id": "some-id",
                        },
                    }
                ],
                "release-group": {
                    "id": "another-id",
                },
                "country": "COUNTRY",
            },
        ]

        with mock.patch(
            "beetsplug._utils.musicbrainz.MusicBrainzAPI.get_release"
        ) as gp:
            gp.side_effect = side_effect
            album = self.mb.album_for_id("d2a6f856-b553-40a0-ac54-a321e8e2da02")
            assert album.country == "COUNTRY"

    def test_pseudo_releases_with_empty_links(self):
        side_effect = [
            {
                "title": "pseudo",
                "id": "d2a6f856-b553-40a0-ac54-a321e8e2da02",
                "status": "Pseudo-Release",
                "media": [
                    {
                        "tracks": [
                            {
                                "id": "baz",
                                "recording": {
                                    "title": "translated title",
                                    "id": "bar",
                                    "length": 42,
                                },
                                "position": 9,
                                "number": "A1",
                            }
                        ],
                        "position": 5,
                    }
                ],
                "artist-credit": [
                    {
                        "artist": {
                            "name": "some-artist",
                            "id": "some-id",
                        },
                    }
                ],
                "release-group": {
                    "id": "another-id",
                },
            }
        ]

        with mock.patch(
            "beetsplug._utils.musicbrainz.MusicBrainzAPI.get_release"
        ) as gp:
            gp.side_effect = side_effect
            album = self.mb.album_for_id("d2a6f856-b553-40a0-ac54-a321e8e2da02")
            assert album.country is None

    def test_pseudo_releases_without_links(self):
        side_effect = [
            {
                "title": "pseudo",
                "id": "d2a6f856-b553-40a0-ac54-a321e8e2da02",
                "status": "Pseudo-Release",
                "media": [
                    {
                        "tracks": [
                            {
                                "id": "baz",
                                "recording": {
                                    "title": "translated title",
                                    "id": "bar",
                                    "length": 42,
                                },
                                "position": 9,
                                "number": "A1",
                            }
                        ],
                        "position": 5,
                    }
                ],
                "artist-credit": [
                    {
                        "artist": {
                            "name": "some-artist",
                            "id": "some-id",
                        },
                    }
                ],
                "release-group": {
                    "id": "another-id",
                },
            }
        ]

        with mock.patch(
            "beetsplug._utils.musicbrainz.MusicBrainzAPI.get_release"
        ) as gp:
            gp.side_effect = side_effect
            album = self.mb.album_for_id("d2a6f856-b553-40a0-ac54-a321e8e2da02")
            assert album.country is None

    def test_pseudo_releases_with_unsupported_links(self):
        side_effect = [
            {
                "title": "pseudo",
                "id": "d2a6f856-b553-40a0-ac54-a321e8e2da02",
                "status": "Pseudo-Release",
                "media": [
                    {
                        "tracks": [
                            {
                                "id": "baz",
                                "recording": {
                                    "title": "translated title",
                                    "id": "bar",
                                    "length": 42,
                                },
                                "position": 9,
                                "number": "A1",
                            }
                        ],
                        "position": 5,
                    }
                ],
                "artist-credit": [
                    {
                        "artist": {
                            "name": "some-artist",
                            "id": "some-id",
                        },
                    }
                ],
                "release-group": {
                    "id": "another-id",
                },
                "release-relations": [
                    {
                        "type": "remaster",
                        "direction": "backward",
                        "release": {
                            "id": "d2a6f856-b553-40a0-ac54-a321e8e2da01"
                        },
                    }
                ],
            }
        ]

        with mock.patch(
            "beetsplug._utils.musicbrainz.MusicBrainzAPI.get_release"
        ) as gp:
            gp.side_effect = side_effect
            album = self.mb.album_for_id("d2a6f856-b553-40a0-ac54-a321e8e2da02")
            assert album.country is None


class TestMusicBrainzPlugin(PluginMixin):
    plugin = "musicbrainz"

    mbid = "d2a6f856-b553-40a0-ac54-a321e8e2da99"
    RECORDING: ClassVar[dict[str, int | str]] = {
        "title": "foo",
        "id": mbid,
        "length": 42,
    }

    @pytest.fixture
    def plugin_config(self):
        return {}

    @pytest.fixture
    def mb(self, plugin_config):
        self.config[self.plugin].set(plugin_config)

        return musicbrainz.MusicBrainzPlugin()

    @pytest.mark.parametrize(
        "plugin_config,va_likely,expected_additional_criteria",
        [
            ({}, False, {"artist": "Artist "}),
            ({}, True, {"arid": "89ad4ac3-39f7-470e-963a-56509c546377"}),
            (
                {"extra_tags": ["label", "catalognum"]},
                False,
                {"artist": "Artist ", "label": "abc", "catno": "ABC123"},
            ),
        ],
    )
    def test_get_album_criteria(
        self, mb, va_likely, expected_additional_criteria
    ):
        items = [
            Item(catalognum="ABC 123", label="abc"),
            Item(catalognum="ABC 123", label="abc"),
            Item(catalognum="ABC 123", label="def"),
        ]

        assert mb.get_album_criteria(items, "Artist ", " Album", va_likely) == {
            "release": " Album",
            **expected_additional_criteria,
        }

    def test_item_candidates(self, monkeypatch, mb):
        monkeypatch.setattr(
            "beetsplug._utils.musicbrainz.MusicBrainzAPI.get_json",
            lambda *_, **__: {"recordings": [self.RECORDING]},
        )
        monkeypatch.setattr(
            "beetsplug._utils.musicbrainz.MusicBrainzAPI.get_recording",
            lambda *_, **__: self.RECORDING,
        )

        candidates = list(mb.item_candidates(Item(), "hello", "there"))

        assert len(candidates) == 1
        assert candidates[0].track_id == self.RECORDING["id"]

    def test_candidates(self, monkeypatch, mb):
        monkeypatch.setattr(
            "beetsplug._utils.musicbrainz.MusicBrainzAPI.get_json",
            lambda *_, **__: {"releases": [{"id": self.mbid}]},
        )
        monkeypatch.setattr(
            "beetsplug._utils.musicbrainz.MusicBrainzAPI.get_release",
            lambda *_, **__: {
                "title": "hi",
                "id": self.mbid,
                "status": "status",
                "media": [
                    {
                        "tracks": [
                            {
                                "id": "baz",
                                "recording": self.RECORDING,
                                "position": 9,
                                "number": "A1",
                            }
                        ],
                        "position": 5,
                    }
                ],
                "artist-credit": [
                    {"artist": {"name": "some-artist", "id": "some-id"}}
                ],
                "release-group": {"id": "another-id"},
            },
        )
        candidates = list(mb.candidates([], "hello", "there", False))

        assert len(candidates) == 1
        assert candidates[0].tracks[0].track_id == self.RECORDING["id"]
        assert candidates[0].album == "hi"

    def test_import_handles_404_gracefully(self, mb, requests_mock):
        id_ = uuid.uuid4()
        response = requests.Response()
        response.status_code = 404
        requests_mock.get(
            f"/ws/2/release/{id_}",
            exc=requests.exceptions.HTTPError(response=response),
        )
        res = mb.album_for_id(str(id_))
        assert res is None

    def test_import_propagates_non_404_errors(self, mb):
        class DummyResponse:
            status_code = 500

        error = requests.exceptions.HTTPError(response=DummyResponse())

        def raise_error(*args, **kwargs):
            raise error

        # Simulate mb.mb_api.get_release raising a non-404 HTTP error
        mb.mb_api.get_release = raise_error

        with pytest.raises(requests.exceptions.HTTPError) as excinfo:
            mb.album_for_id(str(uuid.uuid4()))

        # Ensure the exact error is propagated, not swallowed
        assert excinfo.value is error



# This file is part of beets.
# Copyright 2016, Adrian Sampson.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Updates an MPD index whenever the library is changed.

Put something like the following in your config.yaml to configure:
    mpd:
        host: localhost
        port: 6600
        password: seekrit
"""

import os
import socket

from beets import config
from beets.plugins import BeetsPlugin


# No need to introduce a dependency on an MPD library for such a
# simple use case. Here's a simple socket abstraction to make things
# easier.
class BufferedSocket:
    """Socket abstraction that allows reading by line."""

    def __init__(self, host, port, sep=b"\n"):
        if host[0] in ["/", "~"]:
            self.sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self.sock.connect(os.path.expanduser(host))
        else:
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.sock.connect((host, port))
        self.buf = b""
        self.sep = sep

    def readline(self):
        while self.sep not in self.buf:
            data = self.sock.recv(1024)
            if not data:
                break
            self.buf += data
        if self.sep in self.buf:
            res, self.buf = self.buf.split(self.sep, 1)
            return res + self.sep
        else:
            return b""

    def send(self, data):
        self.sock.send(data)

    def close(self):
        self.sock.close()


class MPDUpdatePlugin(BeetsPlugin):
    def __init__(self):
        super().__init__()
        config["mpd"].add(
            {
                "host": os.environ.get("MPD_HOST", "localhost"),
                "port": int(os.environ.get("MPD_PORT", 6600)),
                "password": "",
            }
        )
        config["mpd"]["password"].redact = True

        # For backwards compatibility, use any values from the
        # plugin-specific "mpdupdate" section.
        for key in config["mpd"].keys():
            if self.config[key].exists():
                config["mpd"][key] = self.config[key].get()

        self.register_listener("database_change", self.db_change)

    def db_change(self, lib, model):
        self.register_listener("cli_exit", self.update)

    def update(self, lib):
        self.update_mpd(
            config["mpd"]["host"].as_str(),
            config["mpd"]["port"].get(int),
            config["mpd"]["password"].as_str(),
        )

    def update_mpd(self, host="localhost", port=6600, password=None):
        """Sends the "update" command to the MPD server indicated,
        possibly authenticating with a password first.
        """
        self._log.info("Updating MPD database...")

        try:
            s = BufferedSocket(host, port)
        except OSError:
            self._log.warning("MPD connection failed", exc_info=True)
            return

        resp = s.readline()
        if b"OK MPD" not in resp:
            self._log.warning("MPD connection failed: {0!r}", resp)
            return

        if password:
            s.send(f'password "{password}"\n'.encode())
            resp = s.readline()
            if b"OK" not in resp:
                self._log.warning("Authentication failed: {0!r}", resp)
                s.send(b"close\n")
                s.close()
                return

        s.send(b"update\n")
        resp = s.readline()
        if b"updating_db" not in resp:
            self._log.warning("Update failed: {0!r}", resp)

        s.send(b"close\n")
        s.close()
        self._log.info("Database updated.")



"""The `modify` command: change metadata fields."""

from beets import library, ui
from beets.util import functemplate

from .utils import do_query


def modify_items(lib, mods, dels, query, write, move, album, confirm, inherit):
    """Modifies matching items according to user-specified assignments and
    deletions.

    `mods` is a dictionary of field and value pairse indicating
    assignments. `dels` is a list of fields to be deleted.
    """
    # Parse key=value specifications into a dictionary.
    model_cls = library.Album if album else library.Item

    # Get the items to modify.
    items, albums = do_query(lib, query, album, False)
    objs = albums if album else items

    # Apply changes *temporarily*, preview them, and collect modified
    # objects.
    ui.print_(f"Modifying {len(objs)} {'album' if album else 'item'}s.")
    changed = []
    templates = {
        key: functemplate.template(value) for key, value in mods.items()
    }
    for obj in objs:
        obj_mods = {
            key: model_cls._parse(key, obj.evaluate_template(templates[key]))
            for key in mods.keys()
        }
        if print_and_modify(obj, obj_mods, dels) and obj not in changed:
            changed.append(obj)

    # Still something to do?
    if not changed:
        ui.print_("No changes to make.")
        return

    # Confirm action.
    if confirm:
        if write and move:
            extra = ", move and write tags"
        elif write:
            extra = " and write tags"
        elif move:
            extra = " and move"
        else:
            extra = ""

        changed = ui.input_select_objects(
            f"Really modify{extra}",
            changed,
            lambda o: print_and_modify(o, mods, dels),
        )

    # Apply changes to database and files
    with lib.transaction():
        for obj in changed:
            obj.try_sync(write, move, inherit)


def print_and_modify(obj, mods, dels):
    """Print the modifications to an item and return a bool indicating
    whether any changes were made.

    `mods` is a dictionary of fields and values to update on the object;
    `dels` is a sequence of fields to delete.
    """
    obj.update(mods)
    for field in dels:
        try:
            del obj[field]
        except KeyError:
            pass
    return ui.show_model_changes(obj)


def modify_parse_args(args):
    """Split the arguments for the modify subcommand into query parts,
    assignments (field=value), and deletions (field!).  Returns the result as
    a three-tuple in that order.
    """
    mods = {}
    dels = []
    query = []
    for arg in args:
        if arg.endswith("!") and "=" not in arg and ":" not in arg:
            dels.append(arg[:-1])  # Strip trailing !.
        elif "=" in arg and ":" not in arg.split("=", 1)[0]:
            key, val = arg.split("=", 1)
            mods[key] = val
        else:
            query.append(arg)
    return query, mods, dels


def modify_func(lib, opts, args):
    query, mods, dels = modify_parse_args(args)
    if not mods and not dels:
        raise ui.UserError("no modifications specified")
    modify_items(
        lib,
        mods,
        dels,
        query,
        ui.should_write(opts.write),
        ui.should_move(opts.move),
        opts.album,
        not opts.yes,
        opts.inherit,
    )


modify_cmd = ui.Subcommand(
    "modify", help="change metadata fields", aliases=("mod",)
)
modify_cmd.parser.add_option(
    "-m",
    "--move",
    action="store_true",
    dest="move",
    help="move files in the library directory",
)
modify_cmd.parser.add_option(
    "-M",
    "--nomove",
    action="store_false",
    dest="move",
    help="don't move files in library",
)
modify_cmd.parser.add_option(
    "-w",
    "--write",
    action="store_true",
    default=None,
    help="write new metadata to files' tags (default)",
)
modify_cmd.parser.add_option(
    "-W",
    "--nowrite",
    action="store_false",
    dest="write",
    help="don't write metadata (opposite of -w)",
)
modify_cmd.parser.add_album_option()
modify_cmd.parser.add_format_option(target="item")
modify_cmd.parser.add_option(
    "-y", "--yes", action="store_true", help="skip confirmation"
)
modify_cmd.parser.add_option(
    "-I",
    "--noinherit",
    action="store_false",
    dest="inherit",
    default=True,
    help="when modifying albums, don't also change item data",
)
modify_cmd.func = modify_func



"""Adds Listenbrainz support to Beets."""

import datetime

import requests

from beets import config, ui
from beets.plugins import BeetsPlugin
from beetsplug.lastimport import process_tracks

from ._utils.musicbrainz import MusicBrainzAPIMixin


class ListenBrainzPlugin(MusicBrainzAPIMixin, BeetsPlugin):
    """A Beets plugin for interacting with ListenBrainz."""

    ROOT = "http://api.listenbrainz.org/1/"

    def __init__(self):
        """Initialize the plugin."""
        super().__init__()
        self.token = self.config["token"].get()
        self.username = self.config["username"].get()
        self.AUTH_HEADER = {"Authorization": f"Token {self.token}"}
        config["listenbrainz"]["token"].redact = True

    def commands(self):
        """Add beet UI commands to interact with ListenBrainz."""
        lbupdate_cmd = ui.Subcommand(
            "lbimport", help="Import ListenBrainz history"
        )

        def func(lib, opts, args):
            self._lbupdate(lib, self._log)

        lbupdate_cmd.func = func
        return [lbupdate_cmd]

    def _lbupdate(self, lib, log):
        """Obtain view count from Listenbrainz."""
        found_total = 0
        unknown_total = 0
        ls = self.get_listens()
        tracks = self.get_tracks_from_listens(ls)
        log.info("Found {} listens", len(ls))
        if tracks:
            found, unknown = process_tracks(lib, tracks, log)
            found_total += found
            unknown_total += unknown
        log.info("... done!")
        log.info("{} unknown play-counts", unknown_total)
        log.info("{} play-counts imported", found_total)

    def _make_request(self, url, params=None):
        """Makes a request to the ListenBrainz API."""
        try:
            response = requests.get(
                url=url,
                headers=self.AUTH_HEADER,
                timeout=10,
                params=params,
            )
            response.raise_for_status()
            return response.json()
        except requests.exceptions.RequestException as e:
            self._log.debug("Invalid Search Error: {}", e)
            return None

    def get_listens(self, min_ts=None, max_ts=None, count=None):
        """Gets the listen history of a given user.

        Args:
            username: User to get listen history of.
            min_ts: History before this timestamp will not be returned.
                    DO NOT USE WITH max_ts.
            max_ts: History after this timestamp will not be returned.
                    DO NOT USE WITH min_ts.
            count: How many listens to return. If not specified,
                uses a default from the server.

        Returns:
            A list of listen info dictionaries if there's an OK status.

        Raises:
            An HTTPError if there's a failure.
            A ValueError if the JSON in the response is invalid.
            An IndexError if the JSON is not structured as expected.
        """
        url = f"{self.ROOT}/user/{self.username}/listens"
        params = {
            k: v
            for k, v in {
                "min_ts": min_ts,
                "max_ts": max_ts,
                "count": count,
            }.items()
            if v is not None
        }
        response = self._make_request(url, params)

        if response is not None:
            return response["payload"]["listens"]
        else:
            return None

    def get_tracks_from_listens(self, listens):
        """Returns a list of tracks from a list of listens."""
        tracks = []
        for track in listens:
            if track["track_metadata"].get("release_name") is None:
                continue
            mbid_mapping = track["track_metadata"].get("mbid_mapping", {})
            mbid = None
            if mbid_mapping.get("recording_mbid") is None:
                # search for the track using title and release
                mbid = self.get_mb_recording_id(track)
            tracks.append(
                {
                    "album": {
                        "name": track["track_metadata"].get("release_name")
                    },
                    "name": track["track_metadata"].get("track_name"),
                    "artist": {
                        "name": track["track_metadata"].get("artist_name")
                    },
                    "mbid": mbid,
                    "release_mbid": mbid_mapping.get("release_mbid"),
                    "listened_at": track.get("listened_at"),
                }
            )
        return tracks

    def get_mb_recording_id(self, track) -> str | None:
        """Returns the MusicBrainz recording ID for a track."""
        results = self.mb_api.search(
            "recording",
            {
                "": track["track_metadata"].get("track_name"),
                "release": track["track_metadata"].get("release_name"),
            },
        )
        return next((r["id"] for r in results), None)

    def get_playlists_createdfor(self, username):
        """Returns a list of playlists created by a user."""
        url = f"{self.ROOT}/user/{username}/playlists/createdfor"
        return self._make_request(url)

    def get_listenbrainz_playlists(self):
        resp = self.get_playlists_createdfor(self.username)
        playlists = resp.get("playlists")
        listenbrainz_playlists = []

        for playlist in playlists:
            playlist_info = playlist.get("playlist")
            if playlist_info.get("creator") == "listenbrainz":
                title = playlist_info.get("title")
                self._log.debug("Playlist title: {}", title)
                playlist_type = (
                    "Exploration" if "Exploration" in title else "Jams"
                )
                if "week of" in title:
                    date_str = title.split("week of ")[1].split(" ")[0]
                    date = datetime.datetime.strptime(
                        date_str, "%Y-%m-%d"
                    ).date()
                else:
                    continue
                identifier = playlist_info.get("identifier")
                id = identifier.split("/")[-1]
                listenbrainz_playlists.append(
                    {"type": playlist_type, "date": date, "identifier": id}
                )
        listenbrainz_playlists = sorted(
            listenbrainz_playlists, key=lambda x: x["type"]
        )
        listenbrainz_playlists = sorted(
            listenbrainz_playlists, key=lambda x: x["date"], reverse=True
        )
        for playlist in listenbrainz_playlists:
            self._log.debug("Playlist: {0[type]} - {0[date]}", playlist)
        return listenbrainz_playlists

    def get_playlist(self, identifier):
        """Returns a playlist."""
        url = f"{self.ROOT}/playlist/{identifier}"
        return self._make_request(url)

    def get_tracks_from_playlist(self, playlist):
        """This function returns a list of tracks in the playlist."""
        tracks = []
        for track in playlist.get("playlist").get("track"):
            identifier = track.get("identifier")
            if isinstance(identifier, list):
                identifier = identifier[0]

            tracks.append(
                {
                    "artist": track.get("creator", "Unknown artist"),
                    "identifier": identifier.split("/")[-1],
                    "title": track.get("title"),
                }
            )
        return self.get_track_info(tracks)

    def get_track_info(self, tracks):
        track_info = []
        for track in tracks:
            identifier = track.get("identifier")
            recording = self.mb_api.get_recording(
                identifier, includes=["releases", "artist-credits"]
            )
            title = recording.get("title")
            artist_credit = recording.get("artist-credit", [])
            if artist_credit:
                artist = artist_credit[0].get("artist", {}).get("name")
            else:
                artist = None
            releases = recording.get("releases", [])
            if releases:
                album = releases[0].get("title")
                date = releases[0].get("date")
                year = date.split("-")[0] if date else None
            else:
                album = None
                year = None
            track_info.append(
                {
                    "identifier": identifier,
                    "title": title,
                    "artist": artist,
                    "album": album,
                    "year": year,
                }
            )
        return track_info

    def get_weekly_playlist(self, playlist_type, most_recent=True):
        # Fetch all playlists
        playlists = self.get_listenbrainz_playlists()
        # Filter playlists by type
        filtered_playlists = [
            p for p in playlists if p["type"] == playlist_type
        ]
        # Sort playlists by date in descending order
        sorted_playlists = sorted(
            filtered_playlists, key=lambda x: x["date"], reverse=True
        )
        # Select the most recent or older playlist based on the most_recent flag
        selected_playlist = (
            sorted_playlists[0] if most_recent else sorted_playlists[1]
        )
        self._log.debug(
            f"Selected playlist: {selected_playlist['type']} "
            f"- {selected_playlist['date']}"
        )
        # Fetch and return tracks from the selected playlist
        playlist = self.get_playlist(selected_playlist.get("identifier"))
        return self.get_tracks_from_playlist(playlist)

    def get_weekly_exploration(self):
        return self.get_weekly_playlist("Exploration", most_recent=True)

    def get_weekly_jams(self):
        return self.get_weekly_playlist("Jams", most_recent=True)

    def get_last_weekly_exploration(self):
        return self.get_weekly_playlist("Exploration", most_recent=False)

    def get_last_weekly_jams(self):
        return self.get_weekly_playlist("Jams", most_recent=False)



from __future__ import annotations

import os
import string
import sys
import time
import unicodedata
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, ClassVar

from mediafile import MediaFile, UnreadableFileError

import beets
from beets import dbcore, logging, plugins, util
from beets.dbcore import types
from beets.util import (
    MoveOperation,
    bytestring_path,
    cached_classproperty,
    normpath,
    samefile,
    syspath,
)
from beets.util.functemplate import Template, template

from .exceptions import FileOperationError, ReadError, WriteError
from .queries import PF_KEY_DEFAULT, parse_query_string

if TYPE_CHECKING:
    from ..dbcore.query import FieldQuery, FieldQueryType
    from .library import Library  # noqa: F401

log = logging.getLogger("beets")


class LibModel(dbcore.Model["Library"]):
    """Shared concrete functionality for Items and Albums."""

    # Config key that specifies how an instance should be formatted.
    _format_config_key: str
    path: bytes
    length: float

    @cached_classproperty
    def _types(cls) -> dict[str, types.Type]:
        """Return the types of the fields in this model."""
        return {
            **plugins.types(cls),  # type: ignore[arg-type]
            "data_source": types.STRING,
        }

    @cached_classproperty
    def _queries(cls) -> dict[str, FieldQueryType]:
        return plugins.named_queries(cls)  # type: ignore[arg-type]

    @cached_classproperty
    def writable_media_fields(cls) -> set[str]:
        return set(MediaFile.fields()) & cls._fields.keys()

    @property
    def filepath(self) -> Path:
        """The path to the entity as pathlib.Path."""
        return Path(os.fsdecode(self.path))

    def _template_funcs(self):
        funcs = DefaultTemplateFunctions(self, self._db).functions()
        funcs.update(plugins.template_funcs())
        return funcs

    def store(self, fields=None):
        super().store(fields)
        plugins.send("database_change", lib=self._db, model=self)

    def remove(self):
        super().remove()
        plugins.send("database_change", lib=self._db, model=self)

    def add(self, lib=None):
        # super().add() calls self.store(), which sends `database_change`,
        # so don't do it here
        super().add(lib)

    def __format__(self, spec):
        if not spec:
            spec = beets.config[self._format_config_key].as_str()
        assert isinstance(spec, str)
        return self.evaluate_template(spec)

    def __str__(self):
        return format(self)

    def __bytes__(self):
        return self.__str__().encode("utf-8")

    # Convenient queries.

    @classmethod
    def field_query(
        cls, field: str, pattern: str, query_cls: FieldQueryType
    ) -> FieldQuery:
        """Get a `FieldQuery` for the given field on this model."""
        fast = field in cls.all_db_fields
        if field in cls.shared_db_fields:
            # This field exists in both tables, so SQLite will encounter
            # an OperationalError if we try to use it in a query.
            # Using an explicit table name resolves this.
            field = f"{cls._table}.{field}"

        return query_cls(field, pattern, fast)

    @classmethod
    def any_field_query(cls, *args, **kwargs) -> dbcore.OrQuery:
        return dbcore.OrQuery(
            [cls.field_query(f, *args, **kwargs) for f in cls._search_fields]
        )

    @classmethod
    def any_writable_media_field_query(cls, *args, **kwargs) -> dbcore.OrQuery:
        fields = cls.writable_media_fields
        return dbcore.OrQuery(
            [cls.field_query(f, *args, **kwargs) for f in fields]
        )

    def duplicates_query(self, fields: list[str]) -> dbcore.AndQuery:
        """Return a query for entities with same values in the given fields."""
        return dbcore.AndQuery(
            [
                self.field_query(f, self.get(f), dbcore.MatchQuery)
                for f in fields
            ]
        )


class FormattedItemMapping(dbcore.db.FormattedMapping):
    """Add lookup for album-level fields.

    Album-level fields take precedence if `for_path` is true.
    """

    ALL_KEYS = "*"

    def __init__(self, item, included_keys=ALL_KEYS, for_path=False):
        # We treat album and item keys specially here,
        # so exclude transitive album keys from the model's keys.
        super().__init__(item, included_keys=[], for_path=for_path)
        self.included_keys = included_keys
        if included_keys == self.ALL_KEYS:
            # Performance note: this triggers a database query.
            self.model_keys = item.keys(computed=True, with_album=False)
        else:
            self.model_keys = included_keys
        self.item = item

    @cached_property
    def all_keys(self):
        return set(self.model_keys).union(self.album_keys)

    @cached_property
    def album_keys(self):
        album_keys = []
        if self.album:
            if self.included_keys == self.ALL_KEYS:
                # Performance note: this triggers a database query.
                for key in self.album.keys(computed=True):
                    if (
                        key in Album.item_keys
                        or key not in self.item._fields.keys()
                    ):
                        album_keys.append(key)
            else:
                album_keys = self.included_keys
        return album_keys

    @property
    def album(self):
        return self.item._cached_album

    def _get(self, key):
        """Get the value for a key, either from the album or the item.

        Raise a KeyError for invalid keys.
        """
        if self.for_path and key in self.album_keys:
            return self._get_formatted(self.album, key)
        elif key in self.model_keys:
            return self._get_formatted(self.model, key)
        elif key in self.album_keys:
            return self._get_formatted(self.album, key)
        else:
            raise KeyError(key)

    def __getitem__(self, key):
        """Get the value for a key.

        `artist` and `albumartist` are fallback values for each other
        when not set.
        """
        value = self._get(key)

        # `artist` and `albumartist` fields fall back to one another.
        # This is helpful in path formats when the album artist is unset
        # on as-is imports.
        try:
            if key == "artist" and not value:
                return self._get("albumartist")
            elif key == "albumartist" and not value:
                return self._get("artist")
        except KeyError:
            pass

        return value

    def __iter__(self):
        return iter(self.all_keys)

    def __len__(self):
        return len(self.all_keys)


class Album(LibModel):
    """Provide access to information about albums stored in a
    library.

    Reflects the library's "albums" table, including album art.
    """

    artpath: bytes

    _table = "albums"
    _flex_table = "album_attributes"
    _always_dirty = True
    _fields: ClassVar[dict[str, types.Type]] = {
        "id": types.PRIMARY_ID,
        "artpath": types.NullPathType(),
        "added": types.DATE,
        "albumartist": types.STRING,
        "albumartist_sort": types.STRING,
        "albumartist_credit": types.STRING,
        "albumartists": types.MULTI_VALUE_DSV,
        "albumartists_sort": types.MULTI_VALUE_DSV,
        "albumartists_credit": types.MULTI_VALUE_DSV,
        "album": types.STRING,
        "genres": types.MULTI_VALUE_DSV,
        "style": types.STRING,
        "discogs_albumid": types.INTEGER,
        "discogs_artistid": types.INTEGER,
        "discogs_labelid": types.INTEGER,
        "year": types.PaddedInt(4),
        "month": types.PaddedInt(2),
        "day": types.PaddedInt(2),
        "disctotal": types.PaddedInt(2),
        "comp": types.BOOLEAN,
        "mb_albumid": types.STRING,
        "mb_albumartistid": types.STRING,
        "mb_albumartistids": types.MULTI_VALUE_DSV,
        "albumtype": types.STRING,
        "albumtypes": types.SEMICOLON_SPACE_DSV,
        "label": types.STRING,
        "barcode": types.STRING,
        "mb_releasegroupid": types.STRING,
        "release_group_title": types.STRING,
        "asin": types.STRING,
        "catalognum": types.STRING,
        "script": types.STRING,
        "language": types.STRING,
        "country": types.STRING,
        "albumstatus": types.STRING,
        "albumdisambig": types.STRING,
        "releasegroupdisambig": types.STRING,
        "rg_album_gain": types.NULL_FLOAT,
        "rg_album_peak": types.NULL_FLOAT,
        "r128_album_gain": types.NULL_FLOAT,
        "original_year": types.PaddedInt(4),
        "original_month": types.PaddedInt(2),
        "original_day": types.PaddedInt(2),
    }

    _search_fields = ("album", "albumartist", "genres")

    @cached_classproperty
    def _types(cls) -> dict[str, types.Type]:
        return {**super()._types, "path": types.PathType()}

    _sorts: ClassVar[dict[str, type[dbcore.query.FieldSort]]] = {
        "albumartist": dbcore.query.SmartArtistSort,
        "artist": dbcore.query.SmartArtistSort,
    }

    # List of keys that are set on an album's items.
    item_keys: ClassVar[list[str]] = [
        "added",
        "albumartist",
        "albumartists",
        "albumartist_sort",
        "albumartists_sort",
        "albumartist_credit",
        "albumartists_credit",
        "album",
        "genres",
        "style",
        "discogs_albumid",
        "discogs_artistid",
        "discogs_labelid",
        "year",
        "month",
        "day",
        "disctotal",
        "comp",
        "mb_albumid",
        "mb_albumartistid",
        "mb_albumartistids",
        "albumtype",
        "albumtypes",
        "label",
        "barcode",
        "mb_releasegroupid",
        "asin",
        "catalognum",
        "script",
        "language",
        "country",
        "albumstatus",
        "albumdisambig",
        "releasegroupdisambig",
        "release_group_title",
        "rg_album_gain",
        "rg_album_peak",
        "r128_album_gain",
        "original_year",
        "original_month",
        "original_day",
    ]

    _format_config_key = "format_album"

    @cached_classproperty
    def _relation(cls) -> type[Item]:
        return Item

    @cached_classproperty
    def relation_join(cls) -> str:
        """Return FROM clause which joins on related album items.

        Use LEFT join to select all albums, including those that do not have
        any items.
        """
        return (
            f"LEFT JOIN {cls._relation._table} "
            f"ON {cls._table}.id = {cls._relation._table}.album_id"
        )

    @property
    def art_filepath(self) -> Path | None:
        """The path to album's cover picture as pathlib.Path."""
        return Path(os.fsdecode(self.artpath)) if self.artpath else None

    @classmethod
    def _getters(cls):
        # In addition to plugin-provided computed fields, also expose
        # the album's directory as `path`.
        getters = plugins.album_field_getters()
        getters["path"] = Album.item_dir
        getters["albumtotal"] = Album._albumtotal
        return getters

    def items(self):
        """Return an iterable over the items associated with this
        album.

        This method conflicts with :meth:`LibModel.items`, which is
        inherited from :meth:`beets.dbcore.Model.items`.
        Since :meth:`Album.items` predates these methods, and is
        likely to be used by plugins, we keep this interface as-is.
        """
        return self._db.items(dbcore.MatchQuery("album_id", self.id))

    def remove(self, delete=False, with_items=True):
        """Remove this album and all its associated items from the
        library.

        If delete, then the items' files are also deleted from disk,
        along with any album art. The directories containing the album are
        also removed (recursively) if empty.

        Set with_items to False to avoid removing the album's items.
        """
        super().remove()

        # Send a 'album_removed' signal to plugins
        plugins.send("album_removed", album=self)

        # Delete art file.
        if delete:
            artpath = self.artpath
            if artpath:
                util.remove(artpath)

        # Remove (and possibly delete) the constituent items.
        if with_items:
            for item in self.items():
                item.remove(delete, False)

    def move_art(self, operation=MoveOperation.MOVE):
        """Move, copy, link or hardlink (depending on `operation`) any
        existing album art so that it remains in the same directory as
        the items.

        `operation` should be an instance of `util.MoveOperation`.
        """
        old_art = self.artpath
        if not old_art:
            return

        if not os.path.exists(syspath(old_art)):
            log.error(
                "removing reference to missing album art file {}",
                util.displayable_path(old_art),
            )
            self.artpath = None
            return

        new_art = self.art_destination(old_art)
        if new_art == old_art:
            return

        new_art = util.unique_path(new_art)
        log.debug(
            "moving album art {} to {}",
            util.displayable_path(old_art),
            util.displayable_path(new_art),
        )
        if operation == MoveOperation.MOVE:
            util.move(old_art, new_art)
            util.prune_dirs(os.path.dirname(old_art), self._db.directory)
        elif operation == MoveOperation.COPY:
            util.copy(old_art, new_art)
        elif operation == MoveOperation.LINK:
            util.link(old_art, new_art)
        elif operation == MoveOperation.HARDLINK:
            util.hardlink(old_art, new_art)
        elif operation == MoveOperation.REFLINK:
            util.reflink(old_art, new_art, fallback=False)
        elif operation == MoveOperation.REFLINK_AUTO:
            util.reflink(old_art, new_art, fallback=True)
        else:
            assert False, "unknown MoveOperation"
        self.artpath = new_art

    def move(self, operation=MoveOperation.MOVE, basedir=None, store=True):
        """Move, copy, link or hardlink (depending on `operation`)
        all items to their destination. Any album art moves along with them.

        `basedir` overrides the library base directory for the destination.

        `operation` should be an instance of `util.MoveOperation`.

        By default, the album is stored to the database, persisting any
        modifications to its metadata. If `store` is `False` however,
        the album is not stored automatically, and it will have to be manually
        stored after invoking this method.
        """
        basedir = basedir or self._db.directory

        # Ensure new metadata is available to items for destination
        # computation.
        if store:
            self.store()

        # Move items.
        items = list(self.items())
        for item in items:
            item.move(operation, basedir=basedir, with_album=False, store=store)

        # Move art.
        self.move_art(operation)
        if store:
            self.store()

    def item_dir(self):
        """Return the directory containing the album's first item,
        provided that such an item exists.
        """
        item = self.items().get()
        if not item:
            raise ValueError(f"empty album for album id {self.id}")
        return os.path.dirname(item.path)

    def _albumtotal(self):
        """Return the total number of tracks on all discs on the album."""
        if self.disctotal == 1 or not beets.config["per_disc_numbering"]:
            return self.items()[0].tracktotal

        counted = []
        total = 0

        for item in self.items():
            if item.disc in counted:
                continue

            total += item.tracktotal
            counted.append(item.disc)

            if len(counted) == self.disctotal:
                break

        return total

    def art_destination(self, image, item_dir=None):
        """Return a path to the destination for the album art image
        for the album.

        `image` is the path of the image that will be
        moved there (used for its extension).

        The path construction uses the existing path of the album's
        items, so the album must contain at least one item or
        item_dir must be provided.
        """
        image = bytestring_path(image)
        item_dir = item_dir or self.item_dir()

        filename_tmpl = template(beets.config["art_filename"].as_str())
        subpath = self.evaluate_template(filename_tmpl, True)
        if beets.config["asciify_paths"]:
            subpath = util.asciify_path(
                subpath, beets.config["path_sep_replace"].as_str()
            )
        subpath = util.sanitize_path(
            subpath, replacements=self._db.replacements
        )
        subpath = bytestring_path(subpath)

        _, ext = os.path.splitext(image)
        dest = os.path.join(item_dir, subpath + ext)

        return bytestring_path(dest)

    def set_art(self, path, copy=True):
        """Set the album's cover art to the image at the given path.

        The image is copied (or moved) into place, replacing any
        existing art.

        Send an 'art_set' event with `self` as the sole argument.
        """
        path = bytestring_path(path)
        oldart = self.artpath
        artdest = self.art_destination(path)

        if oldart and samefile(path, oldart):
            # Art already set.
            return
        elif samefile(path, artdest):
            # Art already in place.
            self.artpath = path
            return

        # Normal operation.
        if oldart == artdest:
            util.remove(oldart)
        artdest = util.unique_path(artdest)
        if copy:
            util.copy(path, artdest)
        else:
            util.move(path, artdest)
        self.artpath = artdest

        plugins.send("art_set", album=self)

    def store(self, fields=None, inherit=True):
        """Update the database with the album information.

        `fields` represents the fields to be stored. If not specified,
        all fields will be.

        The album's tracks are also updated when the `inherit` flag is enabled.
        This applies to fixed attributes as well as flexible ones. The `id`
        attribute of the album will never be inherited.
        """
        # Get modified track fields.
        track_updates = {}
        track_deletes = set()
        for key in self._dirty:
            if inherit:
                if key in self.item_keys:  # is a fixed attribute
                    track_updates[key] = self[key]
                elif key not in self:  # is a fixed or a flexible attribute
                    track_deletes.add(key)
                elif key != "id":  # is a flexible attribute
                    track_updates[key] = self[key]

        with self._db.transaction():
            super().store(fields)
            if track_updates:
                for item in self.items():
                    for key, value in track_updates.items():
                        item[key] = value
                    item.store()
            if track_deletes:
                for item in self.items():
                    for key in track_deletes:
                        if key in item:
                            del item[key]
                    item.store()

    def try_sync(self, write, move, inherit=True):
        """Synchronize the album and its items with the database.
        Optionally, also write any new tags into the files and update
        their paths.

        `write` indicates whether to write tags to the item files, and
        `move` controls whether files (both audio and album art) are
        moved.
        """
        self.store(inherit=inherit)
        for item in self.items():
            item.try_sync(write, move)

    @cached_property
    def length(self) -> float:  # type: ignore[override] # still writable since we override __setattr__
        """Return the total length of all items in this album in seconds."""
        return sum(item.length for item in self.items())


class Item(LibModel):
    """Represent a song or track."""

    album_id: int | None

    _table = "items"
    _flex_table = "item_attributes"
    _fields: ClassVar[dict[str, types.Type]] = {
        "id": types.PRIMARY_ID,
        "path": types.PathType(),
        "album_id": types.FOREIGN_ID,
        "title": types.STRING,
        "artist": types.STRING,
        "artists": types.MULTI_VALUE_DSV,
        "artists_ids": types.MULTI_VALUE_DSV,
        "artist_sort": types.STRING,
        "artists_sort": types.MULTI_VALUE_DSV,
        "artist_credit": types.STRING,
        "artists_credit": types.MULTI_VALUE_DSV,
        "remixer": types.STRING,
        "album": types.STRING,
        "albumartist": types.STRING,
        "albumartists": types.MULTI_VALUE_DSV,
        "albumartist_sort": types.STRING,
        "albumartists_sort": types.MULTI_VALUE_DSV,
        "albumartist_credit": types.STRING,
        "albumartists_credit": types.MULTI_VALUE_DSV,
        "genres": types.MULTI_VALUE_DSV,
        "style": types.STRING,
        "discogs_albumid": types.INTEGER,
        "discogs_artistid": types.INTEGER,
        "discogs_labelid": types.INTEGER,
        "lyricist": types.STRING,
        "composer": types.STRING,
        "composer_sort": types.STRING,
        "work": types.STRING,
        "mb_workid": types.STRING,
        "work_disambig": types.STRING,
        "arranger": types.STRING,
        "grouping": types.STRING,
        "year": types.PaddedInt(4),
        "month": types.PaddedInt(2),
        "day": types.PaddedInt(2),
        "track": types.PaddedInt(2),
        "tracktotal": types.PaddedInt(2),
        "disc": types.PaddedInt(2),
        "disctotal": types.PaddedInt(2),
        "lyrics": types.STRING,
        "comments": types.STRING,
        "bpm": types.INTEGER,
        "comp": types.BOOLEAN,
        "mb_trackid": types.STRING,
        "mb_albumid": types.STRING,
        "mb_artistid": types.STRING,
        "mb_artistids": types.MULTI_VALUE_DSV,
        "mb_albumartistid": types.STRING,
        "mb_albumartistids": types.MULTI_VALUE_DSV,
        "mb_releasetrackid": types.STRING,
        "trackdisambig": types.STRING,
        "albumtype": types.STRING,
        "albumtypes": types.SEMICOLON_SPACE_DSV,
        "label": types.STRING,
        "barcode": types.STRING,
        "acoustid_fingerprint": types.STRING,
        "acoustid_id": types.STRING,
        "mb_releasegroupid": types.STRING,
        "release_group_title": types.STRING,
        "asin": types.STRING,
        "isrc": types.STRING,
        "catalognum": types.STRING,
        "script": types.STRING,
        "language": types.STRING,
        "country": types.STRING,
        "albumstatus": types.STRING,
        "media": types.STRING,
        "albumdisambig": types.STRING,
        "releasegroupdisambig": types.STRING,
        "disctitle": types.STRING,
        "encoder": types.STRING,
        "rg_track_gain": types.NULL_FLOAT,
        "rg_track_peak": types.NULL_FLOAT,
        "rg_album_gain": types.NULL_FLOAT,
        "rg_album_peak": types.NULL_FLOAT,
        "r128_track_gain": types.NULL_FLOAT,
        "r128_album_gain": types.NULL_FLOAT,
        "original_year": types.PaddedInt(4),
        "original_month": types.PaddedInt(2),
        "original_day": types.PaddedInt(2),
        "initial_key": types.MusicalKey(),
        "length": types.DurationType(),
        "bitrate": types.ScaledInt(1000, "kbps"),
        "bitrate_mode": types.STRING,
        "encoder_info": types.STRING,
        "encoder_settings": types.STRING,
        "format": types.STRING,
        "samplerate": types.ScaledInt(1000, "kHz"),
        "bitdepth": types.INTEGER,
        "channels": types.INTEGER,
        "mtime": types.DATE,
        "added": types.DATE,
    }
    _indices = (dbcore.Index("idx_item_album_id", ("album_id",)),)

    _search_fields = (
        "artist",
        "title",
        "comments",
        "album",
        "albumartist",
        "genres",
    )

    # Set of item fields that are backed by `MediaFile` fields.
    # Any kind of field (fixed, flexible, and computed) may be a media
    # field. Only these fields are read from disk in `read` and written in
    # `write`.
    _media_fields = set(MediaFile.readable_fields()).intersection(
        _fields.keys()
    )

    # Set of item fields that are backed by *writable* `MediaFile` tag
    # fields.
    # This excludes fields that represent audio data, such as `bitrate` or
    # `length`.
    _media_tag_fields = set(MediaFile.fields()).intersection(_fields.keys())

    _formatter = FormattedItemMapping

    _sorts: ClassVar[dict[str, type[dbcore.query.FieldSort]]] = {
        "artist": dbcore.query.SmartArtistSort
    }

    @cached_classproperty
    def _queries(cls) -> dict[str, FieldQueryType]:
        return {**super()._queries, "singleton": dbcore.query.SingletonQuery}

    _format_config_key = "format_item"

    # Cached album object. Read-only.
    __album: Album | None = None

    @cached_classproperty
    def _relation(cls) -> type[Album]:
        return Album

    @cached_classproperty
    def relation_join(cls) -> str:
        """Return the FROM clause which includes related albums.

        We need to use a LEFT JOIN here, otherwise items that are not part of
        an album (e.g. singletons) would be left out.
        """
        return (
            f"LEFT JOIN {cls._relation._table} "
            f"ON {cls._table}.album_id = {cls._relation._table}.id"
        )

    @property
    def _cached_album(self):
        """The Album object that this item belongs to, if any, or
        None if the item is a singleton or is not associated with a
        library.
        The instance is cached and refreshed on access.

        DO NOT MODIFY!
        If you want a copy to modify, use :meth:`get_album`.
        """
        if not self.__album and self._db:
            self.__album = self._db.get_album(self)
        elif self.__album:
            self.__album.load()
        return self.__album

    @_cached_album.setter
    def _cached_album(self, album):
        self.__album = album

    @classmethod
    def _getters(cls):
        getters = plugins.item_field_getters()
        getters["singleton"] = lambda i: i.album_id is None
        getters["filesize"] = Item.try_filesize  # In bytes.
        return getters

    def duplicates_query(self, fields: list[str]) -> dbcore.AndQuery:
        """Return a query for entities with same values in the given fields."""
        return super().duplicates_query(fields) & dbcore.query.NoneQuery(
            "album_id"
        )

    @classmethod
    def from_path(cls, path):
        """Create a new item from the media file at the specified path."""
        # Initiate with values that aren't read from files.
        i = cls(album_id=None)
        i.read(path)
        i.mtime = i.current_mtime()  # Initial mtime.
        return i

    def __setitem__(self, key, value):
        """Set the item's value for a standard field or a flexattr."""
        # Encode unicode paths and read buffers.
        if key == "path":
            if isinstance(value, str):
                value = bytestring_path(value)
            elif isinstance(value, types.BLOB_TYPE):
                value = bytes(value)
        elif key == "album_id":
            self._cached_album = None

        changed = super()._setitem(key, value)

        if changed and key in MediaFile.fields():
            self.mtime = 0  # Reset mtime on dirty.

    def __getitem__(self, key):
        """Get the value for a field, falling back to the album if
        necessary.

        Raise a KeyError if the field is not available.
        """
        try:
            return super().__getitem__(key)
        except KeyError:
            if self._cached_album:
                return self._cached_album[key]
            raise

    def __repr__(self):
        # This must not use `with_album=True`, because that might access
        # the database. When debugging, that is not guaranteed to succeed, and
        # can even deadlock due to the database lock.
        return (
            f"{type(self).__name__}"
            f"({', '.join(f'{k}={self[k]!r}' for k in self.keys(with_album=False))})"
        )

    def keys(self, computed=False, with_album=True):
        """Get a list of available field names.

        `with_album` controls whether the album's fields are included.
        """
        keys = super().keys(computed=computed)
        if with_album and self._cached_album:
            keys = set(keys)
            keys.update(self._cached_album.keys(computed=computed))
            keys = list(keys)
        return keys

    def get(self, key, default=None, with_album=True):
        """Get the value for a given key or `default` if it does not
        exist.

        Set `with_album` to false to skip album fallback.
        """
        try:
            return self._get(key, default, raise_=with_album)
        except KeyError:
            if self._cached_album:
                return self._cached_album.get(key, default)
            return default

    def update(self, values):
        """Set all key/value pairs in the mapping.

        If mtime is specified, it is not reset (as it might otherwise be).
        """
        super().update(values)
        if self.mtime == 0 and "mtime" in values:
            self.mtime = values["mtime"]

    def clear(self):
        """Set all key/value pairs to None."""
        for key in self._media_tag_fields:
            setattr(self, key, None)

    def get_album(self):
        """Get the Album object that this item belongs to, if any, or
        None if the item is a singleton or is not associated with a
        library.
        """
        if not self._db:
            return None
        return self._db.get_album(self)

    # Interaction with file metadata.

    def read(self, read_path=None):
        """Read the metadata from the associated file.

        If `read_path` is specified, read metadata from that file
        instead. Update all the properties in `_media_fields`
        from the media file.

        Raise a `ReadError` if the file could not be read.
        """
        if read_path is None:
            read_path = self.path
        else:
            read_path = normpath(read_path)
        try:
            mediafile = MediaFile(syspath(read_path))
        except UnreadableFileError as exc:
            raise ReadError(read_path, exc)

        for key in self._media_fields:
            value = getattr(mediafile, key)
            if isinstance(value, int):
                if value.bit_length() > 63:
                    value = 0
            self[key] = value

        # Database's mtime should now reflect the on-disk value.
        if read_path == self.path:
            self.mtime = self.current_mtime()

        self.path = read_path

    def write(self, path=None, tags=None, id3v23=None):
        """Write the item's metadata to a media file.

        All fields in `_media_fields` are written to disk according to
        the values on this object.

        `path` is the path of the mediafile to write the data to. It
        defaults to the item's path.

        `tags` is a dictionary of additional metadata the should be
        written to the file. (These tags need not be in `_media_fields`.)

        `id3v23` will override the global `id3v23` config option if it is
        set to something other than `None`.

        Can raise either a `ReadError` or a `WriteError`.
        """
        if path is None:
            path = self.path
        else:
            path = normpath(path)

        if id3v23 is None:
            id3v23 = beets.config["id3v23"].get(bool)

        # Get the data to write to the file.
        item_tags = dict(self)
        item_tags = {
            k: v for k, v in item_tags.items() if k in self._media_fields
        }  # Only write media fields.
        if tags is not None:
            item_tags.update(tags)
        plugins.send("write", item=self, path=path, tags=item_tags)

        # Open the file.
        try:
            mediafile = MediaFile(syspath(path), id3v23=id3v23)
        except UnreadableFileError as exc:
            raise ReadError(path, exc)

        # Write the tags to the file.
        mediafile.update(item_tags)
        try:
            mediafile.save()
        except UnreadableFileError as exc:
            raise WriteError(self.path, exc)

        # The file has a new mtime.
        if path == self.path:
            self.mtime = self.current_mtime()
        plugins.send("after_write", item=self, path=path)

    def try_write(self, *args, **kwargs):
        """Call `write()` but catch and log `FileOperationError`
        exceptions.

        Return `False` an exception was caught and `True` otherwise.
        """
        try:
            self.write(*args, **kwargs)
            return True
        except FileOperationError as exc:
            log.error("{}", exc)
            return False

    def try_sync(self, write, move, with_album=True):
        """Synchronize the item with the database and, possibly, update its
        tags on disk and its path (by moving the file).

        `write` indicates whether to write new tags into the file. Similarly,
        `move` controls whether the path should be updated. In the
        latter case, files are *only* moved when they are inside their
        library's directory (if any).

        Similar to calling :meth:`write`, :meth:`move`, and :meth:`store`
        (conditionally).
        """
        if write:
            self.try_write()
        if move:
            # Check whether this file is inside the library directory.
            if self._db and self._db.directory in util.ancestry(self.path):
                log.debug("moving {.filepath} to synchronize path", self)
                self.move(with_album=with_album)
        self.store()

    # Files themselves.

    def move_file(self, dest, operation=MoveOperation.MOVE):
        """Move, copy, link or hardlink the item depending on `operation`,
        updating the path value if the move succeeds.

        If a file exists at `dest`, then it is slightly modified to be unique.

        `operation` should be an instance of `util.MoveOperation`.
        """
        if not util.samefile(self.path, dest):
            dest = util.unique_path(dest)
        if operation == MoveOperation.MOVE:
            plugins.send(
                "before_item_moved",
                item=self,
                source=self.path,
                destination=dest,
            )
            util.move(self.path, dest)
            plugins.send(
                "item_moved", item=self, source=self.path, destination=dest
            )
        elif operation == MoveOperation.COPY:
            util.copy(self.path, dest)
            plugins.send(
                "item_copied", item=self, source=self.path, destination=dest
            )
        elif operation == MoveOperation.LINK:
            util.link(self.path, dest)
            plugins.send(
                "item_linked", item=self, source=self.path, destination=dest
            )
        elif operation == MoveOperation.HARDLINK:
            util.hardlink(self.path, dest)
            plugins.send(
                "item_hardlinked", item=self, source=self.path, destination=dest
            )
        elif operation == MoveOperation.REFLINK:
            util.reflink(self.path, dest, fallback=False)
            plugins.send(
                "item_reflinked", item=self, source=self.path, destination=dest
            )
        elif operation == MoveOperation.REFLINK_AUTO:
            util.reflink(self.path, dest, fallback=True)
            plugins.send(
                "item_reflinked", item=self, source=self.path, destination=dest
            )
        else:
            assert False, "unknown MoveOperation"

        # Either copying or moving succeeded, so update the stored path.
        self.path = dest

    def current_mtime(self):
        """Return the current mtime of the file, rounded to the nearest
        integer.
        """
        return int(os.path.getmtime(syspath(self.path)))

    def try_filesize(self):
        """Get the size of the underlying file in bytes.

        If the file is missing, return 0 (and log a warning).
        """
        try:
            return os.path.getsize(syspath(self.path))
        except (OSError, Exception) as exc:
            log.warning("could not get filesize: {}", exc)
            return 0

    # Model methods.

    def remove(self, delete=False, with_album=True):
        """Remove the item.

        If `delete`, then the associated file is removed from disk.

        If `with_album`, then the item's album (if any) is removed
        if the item was the last in the album.
        """
        super().remove()

        # Remove the album if it is empty.
        if with_album:
            album = self.get_album()
            if album and not album.items():
                album.remove(delete, False)

        # Send a 'item_removed' signal to plugins
        plugins.send("item_removed", item=self)

        # Delete the associated file.
        if delete:
            util.remove(self.path)
            util.prune_dirs(os.path.dirname(self.path), self._db.directory)

        self._db._memotable = {}

    def move(
        self,
        operation=MoveOperation.MOVE,
        basedir=None,
        with_album=True,
        store=True,
    ):
        """Move the item to its designated location within the library
        directory (provided by destination()).

        Subdirectories are created as needed. If the operation succeeds,
        the item's path field is updated to reflect the new location.

        Instead of moving the item it can also be copied, linked or hardlinked
        depending on `operation` which should be an instance of
        `util.MoveOperation`.

        `basedir` overrides the library base directory for the destination.

        If the item is in an album and `with_album` is `True`, the album is
        given an opportunity to move its art.

        By default, the item is stored to the database if it is in the
        database, so any dirty fields prior to the move() call will be written
        as a side effect.
        If `store` is `False` however, the item won't be stored and it will
        have to be manually stored after invoking this method.
        """
        dest = self.destination(basedir=basedir)

        # Create necessary ancestry for the move.
        util.mkdirall(dest)

        # Perform the move and store the change.
        old_path = self.path
        self.move_file(dest, operation)
        if store:
            self.store()

        # If this item is in an album, move its art.
        if with_album:
            album = self.get_album()
            if album:
                album.move_art(operation)
                if store:
                    album.store()

        # Prune vacated directory.
        if operation == MoveOperation.MOVE:
            util.prune_dirs(os.path.dirname(old_path), self._db.directory)

    # Templating.

    def destination(
        self,
        relative_to_libdir=False,
        basedir=None,
        path_formats=None,
    ) -> bytes:
        """Return the path in the library directory designated for the item
        (i.e., where the file ought to be).

        The path is returned as a bytestring. ``basedir`` can override the
        library's base directory for the destination. If ``relative_to_libdir``
        is true, returns just the fragment of the path underneath the library
        base directory.
        """
        basedir = basedir or self.db.directory
        path_formats = path_formats or self.db.path_formats

        # Use a path format based on a query, falling back on the
        # default.
        for query, path_format in path_formats:
            if query == PF_KEY_DEFAULT:
                continue
            query, _ = parse_query_string(query, type(self))
            if query.match(self):
                # The query matches the item! Use the corresponding path
                # format.
                break
        else:
            # No query matched; fall back to default.
            for query, path_format in path_formats:
                if query == PF_KEY_DEFAULT:
                    break
            else:
                assert False, "no default path format"
        if isinstance(path_format, Template):
            subpath_tmpl = path_format
        else:
            subpath_tmpl = template(path_format)

        # Evaluate the selected template.
        subpath = self.evaluate_template(subpath_tmpl, True)

        # Prepare path for output: normalize Unicode characters.
        if sys.platform == "darwin":
            subpath = unicodedata.normalize("NFD", subpath)
        else:
            subpath = unicodedata.normalize("NFC", subpath)

        if beets.config["asciify_paths"]:
            subpath = util.asciify_path(
                subpath, beets.config["path_sep_replace"].as_str()
            )

        lib_path_str, fallback = util.legalize_path(
            subpath, self.db.replacements, self.filepath.suffix
        )
        if fallback:
            # Print an error message if legalization fell back to
            # default replacements because of the maximum length.
            log.warning(
                "Fell back to default replacements when naming "
                "file {}. Configure replacements to avoid lengthening "
                "the filename.",
                subpath,
            )
        lib_path_bytes = util.bytestring_path(lib_path_str)

        if relative_to_libdir:
            return lib_path_bytes

        return normpath(os.path.join(basedir, lib_path_bytes))


def _int_arg(s):
    """Convert a string argument to an integer for use in a template
    function.

    May raise a ValueError.
    """
    return int(s.strip())


class DefaultTemplateFunctions:
    """A container class for the default functions provided to path
    templates.

    These functions are contained in an object to provide
    additional context to the functions -- specifically, the Item being
    evaluated.
    """

    _prefix = "tmpl_"

    @cached_classproperty
    def _func_names(cls) -> list[str]:
        """Names of tmpl_* functions in this class."""
        return [s for s in dir(cls) if s.startswith(cls._prefix)]

    def __init__(self, item=None, lib=None):
        """Parametrize the functions.

        If `item` or `lib` is None, then some functions (namely, ``aunique``)
        will always evaluate to the empty string.
        """
        self.item = item
        self.lib = lib

    def functions(self):
        """Return a dictionary containing the functions defined in this
        object.

        The keys are function names (as exposed in templates)
        and the values are Python functions.
        """
        out = {}
        for key in self._func_names:
            out[key[len(self._prefix) :]] = getattr(self, key)
        return out

    @staticmethod
    def tmpl_lower(s):
        """Convert a string to lower case."""
        return s.lower()

    @staticmethod
    def tmpl_upper(s):
        """Convert a string to upper case."""
        return s.upper()

    @staticmethod
    def tmpl_capitalize(s):
        """Converts to a capitalized string."""
        return s.capitalize()

    @staticmethod
    def tmpl_title(s):
        """Convert a string to title case."""
        return string.capwords(s)

    @staticmethod
    def tmpl_left(s, chars):
        """Get the leftmost characters of a string."""
        return s[0 : _int_arg(chars)]

    @staticmethod
    def tmpl_right(s, chars):
        """Get the rightmost characters of a string."""
        return s[-_int_arg(chars) :]

    @staticmethod
    def tmpl_if(condition, trueval, falseval=""):
        """If ``condition`` is nonempty and nonzero, emit ``trueval``;
        otherwise, emit ``falseval`` (if provided).
        """
        try:
            int_condition = _int_arg(condition)
        except ValueError:
            if condition.lower() == "false":
                return falseval
        else:
            condition = int_condition

        if condition:
            return trueval
        else:
            return falseval

    @staticmethod
    def tmpl_asciify(s):
        """Translate non-ASCII characters to their ASCII equivalents."""
        return util.asciify_path(s, beets.config["path_sep_replace"].as_str())

    @staticmethod
    def tmpl_time(s, fmt):
        """Format a time value using `strftime`."""
        cur_fmt = beets.config["time_format"].as_str()
        return time.strftime(fmt, time.strptime(s, cur_fmt))

    def tmpl_aunique(self, keys=None, disam=None, bracket=None):
        """Generate a string that is guaranteed to be unique among all
        albums in the library who share the same set of keys.

        A fields from "disam" is used in the string if one is sufficient to
        disambiguate the albums. Otherwise, a fallback opaque value is
        used. Both "keys" and "disam" should be given as
        whitespace-separated lists of field names, while "bracket" is a
        pair of characters to be used as brackets surrounding the
        disambiguator or empty to have no brackets.
        """
        # Fast paths: no album, no item or library, or memoized value.
        if not self.item or not self.lib:
            return ""

        if isinstance(self.item, Item):
            album_id = self.item.album_id
        elif isinstance(self.item, Album):
            album_id = self.item.id

        if album_id is None:
            return ""

        memokey = self._tmpl_unique_memokey("aunique", keys, disam, album_id)
        memoval = self.lib._memotable.get(memokey)
        if memoval is not None:
            return memoval

        album = self.lib.get_album(album_id)

        return self._tmpl_unique(
            "aunique",
            keys,
            disam,
            bracket,
            album_id,
            album,
            album.item_keys,
            # Do nothing for singletons.
            lambda a: a is None,
        )

    def tmpl_sunique(self, keys=None, disam=None, bracket=None):
        """Generate a string that is guaranteed to be unique among all
        singletons in the library who share the same set of keys.

        A fields from "disam" is used in the string if one is sufficient to
        disambiguate the albums. Otherwise, a fallback opaque value is
        used. Both "keys" and "disam" should be given as
        whitespace-separated lists of field names, while "bracket" is a
        pair of characters to be used as brackets surrounding the
        disambiguator or empty to have no brackets.
        """
        # Fast paths: no album, no item or library, or memoized value.
        if not self.item or not self.lib:
            return ""

        if isinstance(self.item, Item):
            item_id = self.item.id
        else:
            raise NotImplementedError("sunique is only implemented for items")

        if item_id is None:
            return ""

        return self._tmpl_unique(
            "sunique",
            keys,
            disam,
            bracket,
            item_id,
            self.item,
            Item.all_keys(),
            # Do nothing for non singletons.
            lambda i: i.album_id is not None,
        )

    def _tmpl_unique_memokey(self, name, keys, disam, item_id):
        """Get the memokey for the unique template named "name" for the
        specific parameters.
        """
        return (name, keys, disam, item_id)

    def _tmpl_unique(
        self,
        name,
        keys,
        disam,
        bracket,
        item_id,
        db_item,
        item_keys,
        skip_item,
    ):
        """Generate a string that is guaranteed to be unique among all items of
        the same type as "db_item" who share the same set of keys.

        A field from "disam" is used in the string if one is sufficient to
        disambiguate the items. Otherwise, a fallback opaque value is
        used. Both "keys" and "disam" should be given as
        whitespace-separated lists of field names, while "bracket" is a
        pair of characters to be used as brackets surrounding the
        disambiguator or empty to have no brackets.

        "name" is the name of the templates. It is also the name of the
        configuration section where the default values of the parameters
        are stored.

        "skip_item" is a function that must return True when the template
        should return an empty string.

        "initial_subqueries" is a list of subqueries that should be included
        in the query to find the ambiguous items.
        """
        memokey = self._tmpl_unique_memokey(name, keys, disam, item_id)
        memoval = self.lib._memotable.get(memokey)
        if memoval is not None:
            return memoval

        if skip_item(db_item):
            self.lib._memotable[memokey] = ""
            return ""

        keys = keys or beets.config[name]["keys"].as_str()
        disam = disam or beets.config[name]["disambiguators"].as_str()
        if bracket is None:
            bracket = beets.config[name]["bracket"].as_str()
        keys = keys.split()
        disam = disam.split()

        # Assign a left and right bracket or leave blank if argument is empty.
        if len(bracket) == 2:
            bracket_l = bracket[0]
            bracket_r = bracket[1]
        else:
            bracket_l = ""
            bracket_r = ""

        # Find matching items to disambiguate with.
        query = db_item.duplicates_query(keys)
        ambigous_items = (
            self.lib.items(query)
            if isinstance(db_item, Item)
            else self.lib.albums(query)
        )

        # If there's only one item to matching these details, then do
        # nothing.
        if len(ambigous_items) == 1:
            self.lib._memotable[memokey] = ""
            return ""

        # Find the first disambiguator that distinguishes the items.
        for disambiguator in disam:
            # Get the value for each item for the current field.
            disam_values = {s.get(disambiguator, "") for s in ambigous_items}

            # If the set of unique values is equal to the number of
            # items in the disambiguation set, we're done -- this is
            # sufficient disambiguation.
            if len(disam_values) == len(ambigous_items):
                break
        else:
            # No disambiguator distinguished all fields.
            res = f" {bracket_l}{item_id}{bracket_r}"
            self.lib._memotable[memokey] = res
            return res

        # Flatten disambiguation value into a string.
        disam_value = db_item.formatted(for_path=True).get(disambiguator)

        # Return empty string if disambiguator is empty.
        if disam_value:
            res = f" {bracket_l}{disam_value}{bracket_r}"
        else:
            res = ""

        self.lib._memotable[memokey] = res
        return res

    @staticmethod
    def tmpl_first(s, count=1, skip=0, sep="; ", join_str="; "):
        """Get the item(s) from x to y in a string separated by something
        and join then with something.

        Args:
            s: the string
            count: The number of items included
            skip: The number of items skipped
            sep: the separator
            join_str: the string which will join the items
        """
        skip = int(skip)
        count = skip + int(count)
        return join_str.join(s.split(sep)[skip:count])

    def tmpl_ifdef(self, field, trueval="", falseval=""):
        """If field exists return trueval or the field (default)
        otherwise, emit return falseval (if provided).

        Args:
            field: The name of the field
            trueval: The string if the condition is true
            falseval: The string if the condition is false

        Returns:
            The string, based on condition.
        """
        if field in self.item:
            return trueval if trueval else self.item.formatted().get(field)
        else:
            return falseval



# This file is part of beets.
# Copyright 2016, Adrian Sampson.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""A Web interface to beets."""

import base64
import json
import os
import typing as t

import flask
from flask import jsonify
from unidecode import unidecode
from werkzeug.routing import BaseConverter, PathConverter

import beets.library
from beets import ui, util
from beets.dbcore.query import PathQuery
from beets.plugins import BeetsPlugin

# Type checking hacks

if t.TYPE_CHECKING:

    class LibraryCtx(flask.ctx._AppCtxGlobals):
        lib: beets.library.Library

    g = LibraryCtx()
else:
    from flask import g

# Utilities.


def _rep(obj, expand=False):
    """Get a flat -- i.e., JSON-ish -- representation of a beets Item or
    Album object. For Albums, `expand` dictates whether tracks are
    included.
    """
    out = dict(obj)

    if isinstance(obj, beets.library.Item):
        if app.config.get("INCLUDE_PATHS", False):
            out["path"] = util.displayable_path(out["path"])
        else:
            del out["path"]

        # Filter all bytes attributes and convert them to strings.
        for key, value in out.items():
            if isinstance(out[key], bytes):
                out[key] = base64.b64encode(value).decode("ascii")

        # Get the size (in bytes) of the backing file. This is useful
        # for the Tomahawk resolver API.
        try:
            out["size"] = os.path.getsize(util.syspath(obj.path))
        except OSError:
            out["size"] = 0

        return out

    elif isinstance(obj, beets.library.Album):
        if app.config.get("INCLUDE_PATHS", False):
            out["artpath"] = util.displayable_path(out["artpath"])
        else:
            del out["artpath"]
        if expand:
            out["items"] = [_rep(item) for item in obj.items()]
        return out


def json_generator(items, root, expand=False):
    """Generator that dumps list of beets Items or Albums as JSON

    :param root:  root key for JSON
    :param items: list of :class:`Item` or :class:`Album` to dump
    :param expand: If true every :class:`Album` contains its items in the json
                   representation
    :returns:     generator that yields strings
    """
    yield f'{{"{root}":['
    first = True
    for item in items:
        if first:
            first = False
        else:
            yield ","
        yield json.dumps(_rep(item, expand=expand))
    yield "]}"


def is_expand():
    """Returns whether the current request is for an expanded response."""

    return flask.request.args.get("expand") is not None


def is_delete():
    """Returns whether the current delete request should remove the selected
    files.
    """

    return flask.request.args.get("delete") is not None


def get_method():
    """Returns the HTTP method of the current request."""
    return flask.request.method


def resource(name, patchable=False):
    """Decorates a function to handle RESTful HTTP requests for a resource."""

    def make_responder(retriever):
        def responder(ids):
            entities = [retriever(id) for id in ids]
            entities = [entity for entity in entities if entity]

            if get_method() == "DELETE":
                if app.config.get("READONLY", True):
                    return flask.abort(405)

                for entity in entities:
                    entity.remove(delete=is_delete())

                return flask.make_response(jsonify({"deleted": True}), 200)

            elif get_method() == "PATCH" and patchable:
                if app.config.get("READONLY", True):
                    return flask.abort(405)

                for entity in entities:
                    entity.update(flask.request.get_json())
                    entity.try_sync(True, False)  # write, don't move

                if len(entities) == 1:
                    return flask.jsonify(_rep(entities[0], expand=is_expand()))
                elif entities:
                    return app.response_class(
                        json_generator(entities, root=name),
                        mimetype="application/json",
                    )

            elif get_method() == "GET":
                if len(entities) == 1:
                    return flask.jsonify(_rep(entities[0], expand=is_expand()))
                elif entities:
                    return app.response_class(
                        json_generator(entities, root=name),
                        mimetype="application/json",
                    )
                else:
                    return flask.abort(404)

            else:
                return flask.abort(405)

        responder.__name__ = f"get_{name}"

        return responder

    return make_responder


def resource_query(name, patchable=False):
    """Decorates a function to handle RESTful HTTP queries for resources."""

    def make_responder(query_func):
        def responder(queries):
            entities = query_func(queries)

            if get_method() == "DELETE":
                if app.config.get("READONLY", True):
                    return flask.abort(405)

                for entity in entities:
                    entity.remove(delete=is_delete())

                return flask.make_response(jsonify({"deleted": True}), 200)

            elif get_method() == "PATCH" and patchable:
                if app.config.get("READONLY", True):
                    return flask.abort(405)

                for entity in entities:
                    entity.update(flask.request.get_json())
                    entity.try_sync(True, False)  # write, don't move

                return app.response_class(
                    json_generator(entities, root=name),
                    mimetype="application/json",
                )

            elif get_method() == "GET":
                return app.response_class(
                    json_generator(
                        entities, root="results", expand=is_expand()
                    ),
                    mimetype="application/json",
                )

            else:
                return flask.abort(405)

        responder.__name__ = f"query_{name}"

        return responder

    return make_responder


def resource_list(name):
    """Decorates a function to handle RESTful HTTP request for a list of
    resources.
    """

    def make_responder(list_all):
        def responder():
            return app.response_class(
                json_generator(list_all(), root=name, expand=is_expand()),
                mimetype="application/json",
            )

        responder.__name__ = f"all_{name}"
        return responder

    return make_responder


def _get_unique_table_field_values(model, field, sort_field):
    """retrieve all unique values belonging to a key from a model"""
    if field not in model.all_keys() or sort_field not in model.all_keys():
        raise KeyError
    with g.lib.transaction() as tx:
        rows = tx.query(
            f"SELECT DISTINCT {field} FROM {model._table} ORDER BY {sort_field}"
        )
    return [row[0] for row in rows]


class IdListConverter(BaseConverter):
    """Converts comma separated lists of ids in urls to integer lists."""

    def to_python(self, value):
        ids = []
        for id in value.split(","):
            try:
                ids.append(int(id))
            except ValueError:
                pass
        return ids

    def to_url(self, value):
        return ",".join(str(v) for v in value)


class QueryConverter(PathConverter):
    """Converts slash separated lists of queries in the url to string list."""

    def to_python(self, value):
        queries = value.split("/")
        """Do not do path substitution on regex value tests"""
        return [
            query if "::" in query else query.replace("\\", os.sep)
            for query in queries
        ]

    def to_url(self, value):
        return "/".join([v.replace(os.sep, "\\") for v in value])


class EverythingConverter(PathConverter):
    part_isolating = False
    regex = ".*?"


# Flask setup.

app = flask.Flask(__name__)
app.url_map.converters["idlist"] = IdListConverter
app.url_map.converters["query"] = QueryConverter
app.url_map.converters["everything"] = EverythingConverter


@app.before_request
def before_request():
    g.lib = app.config["lib"]


# Items.


@app.route("/item/<idlist:ids>", methods=["GET", "DELETE", "PATCH"])
@resource("items", patchable=True)
def get_item(id):
    return g.lib.get_item(id)


@app.route("/item/")
@app.route("/item/query/")
@resource_list("items")
def all_items():
    return g.lib.items()


@app.route("/item/<int:item_id>/file")
def item_file(item_id):
    item = g.lib.get_item(item_id)

    item_path = util.syspath(item.path)
    base_filename = os.path.basename(item_path)

    try:
        # Imitate http.server behaviour
        base_filename.encode("latin-1", "strict")
    except UnicodeError:
        safe_filename = unidecode(base_filename)
    else:
        safe_filename = base_filename

    response = flask.send_file(
        item_path, as_attachment=True, download_name=safe_filename
    )
    return response


@app.route("/item/query/<query:queries>", methods=["GET", "DELETE", "PATCH"])
@resource_query("items", patchable=True)
def item_query(queries):
    return g.lib.items(queries)


@app.route("/item/path/<everything:path>")
def item_at_path(path):
    query = PathQuery("path", path.encode("utf-8"))
    item = g.lib.items(query).get()
    if item:
        return flask.jsonify(_rep(item))
    else:
        return flask.abort(404)


@app.route("/item/values/<string:key>")
def item_unique_field_values(key):
    sort_key = flask.request.args.get("sort_key", key)
    try:
        values = _get_unique_table_field_values(
            beets.library.Item, key, sort_key
        )
    except KeyError:
        return flask.abort(404)
    return flask.jsonify(values=values)


# Albums.


@app.route("/album/<idlist:ids>", methods=["GET", "DELETE"])
@resource("albums")
def get_album(id):
    return g.lib.get_album(id)


@app.route("/album/")
@app.route("/album/query/")
@resource_list("albums")
def all_albums():
    return g.lib.albums()


@app.route("/album/query/<query:queries>", methods=["GET", "DELETE"])
@resource_query("albums")
def album_query(queries):
    return g.lib.albums(queries)


@app.route("/album/<int:album_id>/art")
def album_art(album_id):
    album = g.lib.get_album(album_id)
    if album and album.artpath:
        return flask.send_file(album.artpath.decode())
    else:
        return flask.abort(404)


@app.route("/album/values/<string:key>")
def album_unique_field_values(key):
    sort_key = flask.request.args.get("sort_key", key)
    try:
        values = _get_unique_table_field_values(
            beets.library.Album, key, sort_key
        )
    except KeyError:
        return flask.abort(404)
    return flask.jsonify(values=values)


# Artists.


@app.route("/artist/")
def all_artists():
    with g.lib.transaction() as tx:
        rows = tx.query("SELECT DISTINCT albumartist FROM albums")
    all_artists = [row[0] for row in rows]
    return flask.jsonify(artist_names=all_artists)


# Library information.


@app.route("/stats")
def stats():
    with g.lib.transaction() as tx:
        item_rows = tx.query("SELECT COUNT(*) FROM items")
        album_rows = tx.query("SELECT COUNT(*) FROM albums")
    return flask.jsonify(
        {
            "items": item_rows[0][0],
            "albums": album_rows[0][0],
        }
    )


# UI.


@app.route("/")
def home():
    return flask.render_template("index.html")


# Plugin hook.


class WebPlugin(BeetsPlugin):
    def __init__(self):
        super().__init__()
        self.config.add(
            {
                "host": "127.0.0.1",
                "port": 8337,
                "cors": "",
                "cors_supports_credentials": False,
                "reverse_proxy": False,
                "include_paths": False,
                "readonly": True,
            }
        )

    def commands(self):
        cmd = ui.Subcommand("web", help="start a Web interface")
        cmd.parser.add_option(
            "-d",
            "--debug",
            action="store_true",
            default=False,
            help="debug mode",
        )

        def func(lib, opts, args):
            args = args
            if args:
                self.config["host"] = args.pop(0)
            if args:
                self.config["port"] = int(args.pop(0))

            app.config["lib"] = lib
            # Normalizes json output
            app.config["JSONIFY_PRETTYPRINT_REGULAR"] = False

            app.config["INCLUDE_PATHS"] = self.config["include_paths"]
            app.config["READONLY"] = self.config["readonly"]

            # Enable CORS if required.
            if self.config["cors"]:
                self._log.info(
                    "Enabling CORS with origin: {}", self.config["cors"]
                )
                from flask_cors import CORS

                app.config["CORS_ALLOW_HEADERS"] = "Content-Type"
                app.config["CORS_RESOURCES"] = {
                    r"/*": {"origins": self.config["cors"].get(str)}
                }
                CORS(
                    app,
                    supports_credentials=self.config[
                        "cors_supports_credentials"
                    ].get(bool),
                )

            # Allow serving behind a reverse proxy
            if self.config["reverse_proxy"]:
                app.wsgi_app = ReverseProxied(app.wsgi_app)

            # Start the web application.
            app.run(
                host=self.config["host"].as_str(),
                port=self.config["port"].get(int),
                debug=opts.debug,
                threaded=True,
            )

        cmd.func = func
        return [cmd]


class ReverseProxied:
    """Wrap the application in this middleware and configure the
    front-end server to add these headers, to let you quietly bind
    this to a URL other than / and to an HTTP scheme that is
    different than what is used locally.

    In nginx:
    location /myprefix {
        proxy_pass http://192.168.0.1:5001;
        proxy_set_header Host $host;
        proxy_set_header X-Forwarded-For $proxy_add_x_forwarded_for;
        proxy_set_header X-Scheme $scheme;
        proxy_set_header X-Script-Name /myprefix;
        }

    From: http://flask.pocoo.org/snippets/35/

    :param app: the WSGI application
    """

    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        script_name = environ.get("HTTP_X_SCRIPT_NAME", "")
        if script_name:
            environ["SCRIPT_NAME"] = script_name
            path_info = environ["PATH_INFO"]
            if path_info.startswith(script_name):
                environ["PATH_INFO"] = path_info[len(script_name) :]

        scheme = environ.get("HTTP_X_SCHEME", "")
        if scheme:
            environ["wsgi.url_scheme"] = scheme
        return self.app(environ, start_response)



# This file is part of beets.
# Copyright 2016, Adrian Sampson.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

"""Tests for the 'beatport' plugin."""

import unittest
from datetime import timedelta

from beets.test import _common
from beets.test.helper import BeetsTestCase
from beetsplug import beatport


class BeatportTest(BeetsTestCase):
    def _make_release_response(self):
        """Returns a dict that mimics a response from the beatport API.

        The results were retrieved from:
        https://oauth-api.beatport.com/catalog/3/releases?id=1742984
        The list of elements on the returned dict is incomplete, including just
        those required for the tests on this class.
        """
        results = {
            "id": 1742984,
            "type": "release",
            "name": "Charade",
            "slug": "charade",
            "releaseDate": "2016-04-11",
            "publishDate": "2016-04-11",
            "audioFormat": "",
            "category": "Release",
            "currentStatus": "General Content",
            "catalogNumber": "GR089",
            "description": "",
            "label": {
                "id": 24539,
                "name": "Gravitas Recordings",
                "type": "label",
                "slug": "gravitas-recordings",
            },
            "artists": [
                {
                    "id": 326158,
                    "name": "Supersillyus",
                    "slug": "supersillyus",
                    "type": "artist",
                }
            ],
            "genres": [
                {"id": 9, "name": "Breaks", "slug": "breaks", "type": "genre"}
            ],
        }
        return results

    def _make_tracks_response(self):
        """Return a list that mimics a response from the beatport API.

        The results were retrieved from:
        https://oauth-api.beatport.com/catalog/3/tracks?releaseId=1742984
        The list of elements on the returned list is incomplete, including just
        those required for the tests on this class.
        """
        results = [
            {
                "id": 7817567,
                "type": "track",
                "sku": "track-7817567",
                "name": "Mirage a Trois",
                "trackNumber": 1,
                "mixName": "Original Mix",
                "title": "Mirage a Trois (Original Mix)",
                "slug": "mirage-a-trois-original-mix",
                "releaseDate": "2016-04-11",
                "publishDate": "2016-04-11",
                "currentStatus": "General Content",
                "length": "7:05",
                "lengthMs": 425421,
                "bpm": 90,
                "key": {
                    "standard": {
                        "letter": "G",
                        "sharp": False,
                        "flat": False,
                        "chord": "minor",
                    },
                    "shortName": "Gmin",
                },
                "artists": [
                    {
                        "id": 326158,
                        "name": "Supersillyus",
                        "slug": "supersillyus",
                        "type": "artist",
                    }
                ],
                "genres": [
                    {
                        "id": 9,
                        "name": "Breaks",
                        "slug": "breaks",
                        "type": "genre",
                    }
                ],
                "subGenres": [
                    {
                        "id": 209,
                        "name": "Glitch Hop",
                        "slug": "glitch-hop",
                        "type": "subgenre",
                    }
                ],
                "release": {
                    "id": 1742984,
                    "name": "Charade",
                    "type": "release",
                    "slug": "charade",
                },
                "label": {
                    "id": 24539,
                    "name": "Gravitas Recordings",
                    "type": "label",
                    "slug": "gravitas-recordings",
                    "status": True,
                },
            },
            {
                "id": 7817568,
                "type": "track",
                "sku": "track-7817568",
                "name": "Aeon Bahamut",
                "trackNumber": 2,
                "mixName": "Original Mix",
                "title": "Aeon Bahamut (Original Mix)",
                "slug": "aeon-bahamut-original-mix",
                "releaseDate": "2016-04-11",
                "publishDate": "2016-04-11",
                "currentStatus": "General Content",
                "length": "7:38",
                "lengthMs": 458000,
                "bpm": 100,
                "key": {
                    "standard": {
                        "letter": "G",
                        "sharp": False,
                        "flat": False,
                        "chord": "major",
                    },
                    "shortName": "Gmaj",
                },
                "artists": [
                    {
                        "id": 326158,
                        "name": "Supersillyus",
                        "slug": "supersillyus",
                        "type": "artist",
                    }
                ],
                "genres": [
                    {
                        "id": 9,
                        "name": "Breaks",
                        "slug": "breaks",
                        "type": "genre",
                    }
                ],
                "subGenres": [
                    {
                        "id": 209,
                        "name": "Glitch Hop",
                        "slug": "glitch-hop",
                        "type": "subgenre",
                    }
                ],
                "release": {
                    "id": 1742984,
                    "name": "Charade",
                    "type": "release",
                    "slug": "charade",
                },
                "label": {
                    "id": 24539,
                    "name": "Gravitas Recordings",
                    "type": "label",
                    "slug": "gravitas-recordings",
                    "status": True,
                },
            },
            {
                "id": 7817569,
                "type": "track",
                "sku": "track-7817569",
                "name": "Trancendental Medication",
                "trackNumber": 3,
                "mixName": "Original Mix",
                "title": "Trancendental Medication (Original Mix)",
                "slug": "trancendental-medication-original-mix",
                "releaseDate": "2016-04-11",
                "publishDate": "2016-04-11",
                "currentStatus": "General Content",
                "length": "1:08",
                "lengthMs": 68571,
                "bpm": 141,
                "key": {
                    "standard": {
                        "letter": "F",
                        "sharp": False,
                        "flat": False,
                        "chord": "major",
                    },
                    "shortName": "Fmaj",
                },
                "artists": [
                    {
                        "id": 326158,
                        "name": "Supersillyus",
                        "slug": "supersillyus",
                        "type": "artist",
                    }
                ],
                "genres": [
                    {
                        "id": 9,
                        "name": "Breaks",
                        "slug": "breaks",
                        "type": "genre",
                    }
                ],
                "subGenres": [
                    {
                        "id": 209,
                        "name": "Glitch Hop",
                        "slug": "glitch-hop",
                        "type": "subgenre",
                    }
                ],
                "release": {
                    "id": 1742984,
                    "name": "Charade",
                    "type": "release",
                    "slug": "charade",
                },
                "label": {
                    "id": 24539,
                    "name": "Gravitas Recordings",
                    "type": "label",
                    "slug": "gravitas-recordings",
                    "status": True,
                },
            },
            {
                "id": 7817570,
                "type": "track",
                "sku": "track-7817570",
                "name": "A List of Instructions for When I'm Human",
                "trackNumber": 4,
                "mixName": "Original Mix",
                "title": "A List of Instructions for When I'm Human (Original Mix)",
                "slug": "a-list-of-instructions-for-when-im-human-original-mix",
                "releaseDate": "2016-04-11",
                "publishDate": "2016-04-11",
                "currentStatus": "General Content",
                "length": "6:57",
                "lengthMs": 417913,
                "bpm": 88,
                "key": {
                    "standard": {
                        "letter": "A",
                        "sharp": False,
                        "flat": False,
                        "chord": "minor",
                    },
                    "shortName": "Amin",
                },
                "artists": [
                    {
                        "id": 326158,
                        "name": "Supersillyus",
                        "slug": "supersillyus",
                        "type": "artist",
                    }
                ],
                "genres": [
                    {
                        "id": 9,
                        "name": "Breaks",
                        "slug": "breaks",
                        "type": "genre",
                    }
                ],
                "subGenres": [
                    {
                        "id": 209,
                        "name": "Glitch Hop",
                        "slug": "glitch-hop",
                        "type": "subgenre",
                    }
                ],
                "release": {
                    "id": 1742984,
                    "name": "Charade",
                    "type": "release",
                    "slug": "charade",
                },
                "label": {
                    "id": 24539,
                    "name": "Gravitas Recordings",
                    "type": "label",
                    "slug": "gravitas-recordings",
                    "status": True,
                },
            },
            {
                "id": 7817571,
                "type": "track",
                "sku": "track-7817571",
                "name": "The Great Shenanigan",
                "trackNumber": 5,
                "mixName": "Original Mix",
                "title": "The Great Shenanigan (Original Mix)",
                "slug": "the-great-shenanigan-original-mix",
                "releaseDate": "2016-04-11",
                "publishDate": "2016-04-11",
                "currentStatus": "General Content",
                "length": "9:49",
                "lengthMs": 589875,
                "bpm": 123,
                "key": {
                    "standard": {
                        "letter": "E",
                        "sharp": False,
                        "flat": True,
                        "chord": "major",
                    },
                    "shortName": "E&#9837;maj",
                },
                "artists": [
                    {
                        "id": 326158,
                        "name": "Supersillyus",
                        "slug": "supersillyus",
                        "type": "artist",
                    }
                ],
                "genres": [
                    {
                        "id": 9,
                        "name": "Breaks",
                        "slug": "breaks",
                        "type": "genre",
                    }
                ],
                "subGenres": [
                    {
                        "id": 209,
                        "name": "Glitch Hop",
                        "slug": "glitch-hop",
                        "type": "subgenre",
                    }
                ],
                "release": {
                    "id": 1742984,
                    "name": "Charade",
                    "type": "release",
                    "slug": "charade",
                },
                "label": {
                    "id": 24539,
                    "name": "Gravitas Recordings",
                    "type": "label",
                    "slug": "gravitas-recordings",
                    "status": True,
                },
            },
            {
                "id": 7817572,
                "type": "track",
                "sku": "track-7817572",
                "name": "Charade",
                "trackNumber": 6,
                "mixName": "Original Mix",
                "title": "Charade (Original Mix)",
                "slug": "charade-original-mix",
                "releaseDate": "2016-04-11",
                "publishDate": "2016-04-11",
                "currentStatus": "General Content",
                "length": "7:05",
                "lengthMs": 425423,
                "bpm": 123,
                "key": {
                    "standard": {
                        "letter": "A",
                        "sharp": False,
                        "flat": False,
                        "chord": "major",
                    },
                    "shortName": "Amaj",
                },
                "artists": [
                    {
                        "id": 326158,
                        "name": "Supersillyus",
                        "slug": "supersillyus",
                        "type": "artist",
                    }
                ],
                "genres": [
                    {
                        "id": 9,
                        "name": "Breaks",
                        "slug": "breaks",
                        "type": "genre",
                    }
                ],
                "subGenres": [
                    {
                        "id": 209,
                        "name": "Glitch Hop",
                        "slug": "glitch-hop",
                        "type": "subgenre",
                    }
                ],
                "release": {
                    "id": 1742984,
                    "name": "Charade",
                    "type": "release",
                    "slug": "charade",
                },
                "label": {
                    "id": 24539,
                    "name": "Gravitas Recordings",
                    "type": "label",
                    "slug": "gravitas-recordings",
                    "status": True,
                },
            },
        ]
        return results

    def setUp(self):
        super().setUp()

        # Set up 'album'.
        response_release = self._make_release_response()
        self.album = beatport.BeatportRelease(response_release)

        # Set up 'tracks'.
        response_tracks = self._make_tracks_response()
        self.tracks = [beatport.BeatportTrack(t) for t in response_tracks]

        # Set up 'test_album'.
        self.test_album = self.mk_test_album()

        # Set up 'test_tracks'
        self.test_tracks = self.test_album.items()

    def mk_test_album(self):
        items = [_common.item() for _ in range(6)]
        for item in items:
            item.album = "Charade"
            item.catalognum = "GR089"
            item.label = "Gravitas Recordings"
            item.artist = "Supersillyus"
            item.year = 2016
            item.comp = False
            item.label_name = "Gravitas Recordings"
            item.genres = ["Glitch Hop", "Breaks"]
            item.year = 2016
            item.month = 4
            item.day = 11
            item.mix_name = "Original Mix"

        items[0].title = "Mirage a Trois"
        items[1].title = "Aeon Bahamut"
        items[2].title = "Trancendental Medication"
        items[3].title = "A List of Instructions for When I'm Human"
        items[4].title = "The Great Shenanigan"
        items[5].title = "Charade"

        items[0].length = timedelta(minutes=7, seconds=5).total_seconds()
        items[1].length = timedelta(minutes=7, seconds=38).total_seconds()
        items[2].length = timedelta(minutes=1, seconds=8).total_seconds()
        items[3].length = timedelta(minutes=6, seconds=57).total_seconds()
        items[4].length = timedelta(minutes=9, seconds=49).total_seconds()
        items[5].length = timedelta(minutes=7, seconds=5).total_seconds()

        items[0].url = "mirage-a-trois-original-mix"
        items[1].url = "aeon-bahamut-original-mix"
        items[2].url = "trancendental-medication-original-mix"
        items[3].url = "a-list-of-instructions-for-when-im-human-original-mix"
        items[4].url = "the-great-shenanigan-original-mix"
        items[5].url = "charade-original-mix"

        counter = 0
        for item in items:
            counter += 1
            item.track_number = counter

        items[0].bpm = 90
        items[1].bpm = 100
        items[2].bpm = 141
        items[3].bpm = 88
        items[4].bpm = 123
        items[5].bpm = 123

        items[0].initial_key = "Gmin"
        items[1].initial_key = "Gmaj"
        items[2].initial_key = "Fmaj"
        items[3].initial_key = "Amin"
        items[4].initial_key = "E&#9837;maj"
        items[5].initial_key = "Amaj"

        for item in items:
            self.lib.add(item)

        album = self.lib.add_album(items)
        album.store()

        return album

    # Test BeatportRelease.
    def test_album_name_applied(self):
        assert self.album.name == self.test_album["album"]

    def test_catalog_number_applied(self):
        assert self.album.catalog_number == self.test_album["catalognum"]

    def test_label_applied(self):
        assert self.album.label_name == self.test_album["label"]

    def test_category_applied(self):
        assert self.album.category == "Release"

    def test_album_url_applied(self):
        assert self.album.url == "https://beatport.com/release/charade/1742984"

    # Test BeatportTrack.
    def test_title_applied(self):
        for track, test_track in zip(self.tracks, self.test_tracks):
            assert track.name == test_track.title

    def test_mix_name_applied(self):
        for track, test_track in zip(self.tracks, self.test_tracks):
            assert track.mix_name == test_track.mix_name

    def test_length_applied(self):
        for track, test_track in zip(self.tracks, self.test_tracks):
            assert int(track.length.total_seconds()) == int(test_track.length)

    def test_track_url_applied(self):
        # Specify beatport ids here because an 'item.id' is beets-internal.
        ids = [
            7817567,
            7817568,
            7817569,
            7817570,
            7817571,
            7817572,
        ]
        # Concatenate with 'id' to pass strict equality test.
        for track, test_track, id in zip(self.tracks, self.test_tracks, ids):
            assert (
                track.url == f"https://beatport.com/track/{test_track.url}/{id}"
            )

    def test_bpm_applied(self):
        for track, test_track in zip(self.tracks, self.test_tracks):
            assert track.bpm == test_track.bpm

    def test_initial_key_applied(self):
        for track, test_track in zip(self.tracks, self.test_tracks):
            assert track.initial_key == test_track.initial_key

    def test_genre_applied(self):
        for track, test_track in zip(self.tracks, self.test_tracks):
            assert track.genres == test_track.genres


class BeatportResponseEmptyTest(unittest.TestCase):
    def _make_tracks_response(self):
        results = [
            {
                "id": 7817567,
                "name": "Mirage a Trois",
                "genres": [
                    {
                        "id": 9,
                        "name": "Breaks",
                        "slug": "breaks",
                        "type": "genre",
                    }
                ],
                "subGenres": [
                    {
                        "id": 209,
                        "name": "Glitch Hop",
                        "slug": "glitch-hop",
                        "type": "subgenre",
                    }
                ],
            }
        ]
        return results

    def setUp(self):
        super().setUp()

        # Set up 'tracks'.
        self.response_tracks = self._make_tracks_response()
        self.tracks = [beatport.BeatportTrack(t) for t in self.response_tracks]

        # Make alias to be congruent with class `BeatportTest`.
        self.test_tracks = self.response_tracks

    def test_response_tracks_empty(self):
        response_tracks = []
        tracks = [beatport.BeatportTrack(t) for t in response_tracks]
        assert tracks == []

    def test_sub_genre_empty_fallback(self):
        """No 'sub_genre' is provided. Test if fallback to 'genre' works."""
        self.response_tracks[0]["subGenres"] = []
        tracks = [beatport.BeatportTrack(t) for t in self.response_tracks]

        self.test_tracks[0]["subGenres"] = []

        assert tracks[0].genres == [self.test_tracks[0]["genres"][0]["name"]]

    def test_genre_empty(self):
        """No 'genre' is provided. Test if 'sub_genre' is applied."""
        self.response_tracks[0]["genres"] = []
        tracks = [beatport.BeatportTrack(t) for t in self.response_tracks]

        self.test_tracks[0]["genres"] = []

        assert tracks[0].genres == [self.test_tracks[0]["subGenres"][0]["name"]]



# This file is part of beets.
# Copyright 2016
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.


from typing import Any, ClassVar
from unittest.mock import ANY, Mock, call, patch

from beets import util
from beets.library import Item
from beets.test.helper import PluginTestCase
from beetsplug.mpdstats import MPDStats


class MPDStatsTest(PluginTestCase):
    plugin = "mpdstats"

    def test_update_rating(self):
        item = Item(title="title", path="", id=1)
        item.add(self.lib)

        log = Mock()
        mpdstats = MPDStats(self.lib, log)

        assert not mpdstats.update_rating(item, True)
        assert not mpdstats.update_rating(None, True)

    def test_get_item(self):
        item_path = util.normpath("/foo/bar.flac")
        item = Item(title="title", path=item_path, id=1)
        item.add(self.lib)

        log = Mock()
        mpdstats = MPDStats(self.lib, log)

        assert str(mpdstats.get_item(item_path)) == str(item)
        assert mpdstats.get_item("/some/non-existing/path") is None
        assert "item not found:" in log.info.call_args[0][0]

    STATUSES: ClassVar[list[dict[str, Any]]] = [
        {"state": "some-unknown-one"},
        {"state": "pause"},
        {"state": "play", "songid": 1, "time": "0:1"},
        {"state": "stop"},
    ]
    EVENTS = [["player"]] * (len(STATUSES) - 1) + [KeyboardInterrupt]
    item_path = util.normpath("/foo/bar.flac")
    songid = 1

    @patch(
        "beetsplug.mpdstats.MPDClientWrapper",
        return_value=Mock(
            **{
                "events.side_effect": EVENTS,
                "status.side_effect": STATUSES,
                "currentsong.return_value": (item_path, songid),
            }
        ),
    )
    def test_run_mpdstats(self, mpd_mock):
        item = Item(title="title", path=self.item_path, id=1)
        item.add(self.lib)

        log = Mock()
        try:
            MPDStats(self.lib, log).run()
        except KeyboardInterrupt:
            pass

        log.debug.assert_has_calls([call('unhandled status "{}"', ANY)])
        log.info.assert_has_calls(
            [call("pause"), call("playing {}", ANY), call("stop")]
        )
